"""
Elastisearch similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

from fiftyone import ViewField as F
import fiftyone.core.utils as fou
import fiftyone.brain.internal.core.utils as fbu
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)

es = fou.lazy_import("elasticsearch")


logger = logging.getLogger(__name__)

_SUPPORTED_METRICS = {
    "cosine": "cosine",
    "dotproduct": "dot_product",
    "euclidean": "l2_norm",
    "innerproduct": "max_inner_product",
}


class ElasticsearchSimilarityConfig(SimilarityConfig):
    """Configuration for a Elasticsearch similarity instance.

    Args:
        index_name (None): the name of the Elasticsearch index to use or
            create. If none is provided, a new index will be created
        metric ("cosine"): the embedding distance metric to use when creating a
            new index. Supported values are
            ``("cosine", "dotproduct", "euclidean", "innerproduct")``
        hosts (None): the full Elasticsearch server address(es) to use. Can be
            a string or list of strings
        cloud_id (None): the Cloud ID of an Elastic Cloud to connect to
        username (None): a username to use
        password (None): a password to use
        api_key (None): an API key to use
        ca_certs (None): a path to a CA certificate
        bearer_auth (None): a bearer token to use
        ssl_assert_fingerprint (None): a SHA256 fingerprint to use
        verify_certs (None): whether to verify SSL certificates
        **kwargs: keyword arguments for
            :class:`fiftyone.brain.similarity.SimilarityConfig`
    """

    def __init__(
        self,
        index_name=None,
        metric="cosine",
        hosts=None,
        cloud_id=None,
        username=None,
        password=None,
        api_key=None,
        ca_certs=None,
        bearer_auth=None,
        ssl_assert_fingerprint=None,
        verify_certs=None,
        **kwargs,
    ):
        if metric not in _SUPPORTED_METRICS:
            raise ValueError(
                "Unsupported metric '%s'. Supported values are %s"
                % (metric, tuple(_SUPPORTED_METRICS.keys()))
            )

        super().__init__(**kwargs)

        self.index_name = index_name
        self.metric = metric

        self._hosts = hosts
        self._cloud_id = cloud_id
        self._username = username
        self._password = password
        self._api_key = api_key
        self._ca_certs = ca_certs
        self._bearer_auth = bearer_auth
        self._ssl_assert_fingerprint = ssl_assert_fingerprint
        self._verify_certs = verify_certs

    @property
    def method(self):
        return "elasticsearch"

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, value):
        self._hosts = value

    @property
    def cloud_id(self):
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, value):
        self._cloud_id = value

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        self._api_key = value

    @property
    def ca_certs(self):
        return self._ca_certs

    @ca_certs.setter
    def ca_certs(self, value):
        self._ca_certs = value

    @property
    def bearer_auth(self):
        return self._bearer_auth

    @bearer_auth.setter
    def bearer_auth(self, value):
        self._bearer_auth = value

    @property
    def ssl_assert_fingerprint(self):
        return self._ssl_assert_fingerprint

    @ssl_assert_fingerprint.setter
    def ssl_assert_fingerprint(self, value):
        self._ssl_assert_fingerprint = value

    @property
    def verify_certs(self):
        return self._verify_certs

    @verify_certs.setter
    def verify_certs(self, value):
        self._verify_certs = value

    @property
    def max_k(self):
        return 10000  # Elasticsearch limit

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    def load_credentials(
        self,
        hosts=None,
        cloud_id=None,
        username=None,
        password=None,
        api_key=None,
        ca_certs=None,
        bearer_auth=None,
        ssl_assert_fingerprint=None,
        verify_certs=None,
    ):
        self._load_parameters(
            hosts=hosts,
            cloud_id=cloud_id,
            username=username,
            password=password,
            api_key=api_key,
            ca_certs=ca_certs,
            bearer_auth=bearer_auth,
            ssl_assert_fingerprint=ssl_assert_fingerprint,
            verify_certs=verify_certs,
        )


class ElasticsearchSimilarity(Similarity):
    """Elasticsearch similarity factory.

    Args:
        config: a :class:`ElasticsearchSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("elasticsearch")

    def ensure_usage_requirements(self):
        fou.ensure_package("elasticsearch")

    def initialize(self, samples, brain_key):
        return ElasticsearchSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class ElasticsearchSimilarityIndex(SimilarityIndex):
    """Class for interacting with Elasticsearch similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`ElasticsearchSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`ElasticsearchSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._client = None
        self._metric = None
        self._initialize()

    @property
    def total_index_size(self):
        try:
            return self._client.count(index=self.config.index_name)["count"]
        except:
            return 0

    @property
    def client(self):
        """The ``elasticsearch.Elasticsearch`` instance for this index."""
        return self._client

    def _initialize(self):
        kwargs = {}

        for key in (
            "hosts",
            "cloud_id",
            "username",
            "password",
            "api_key",
            "ca_certs",
            "bearer_auth",
            "ssl_assert_fingerprint",
            "verify_certs",
        ):
            value = getattr(self.config, key, None)
            if value is not None:
                kwargs[key] = value

        username = kwargs.pop("username", None)
        password = kwargs.pop("password", None)
        if username is not None and password is not None:
            kwargs["basic_auth"] = (username, password)

        try:
            self._client = es.Elasticsearch(**kwargs)
        except Exception as e:
            raise ValueError(
                "Failed to connect to Elasticsearch backend. Refer to "
                "https://docs.voxel51.com/integrations/elasticsearch.html for more "
                "information"
            ) from e

        if self.config.index_name is None:
            root = "fiftyone-" + fou.to_slug(self.samples._root_dataset.name)
            index_name = fbu.get_unique_name(root, self._get_index_names())

            self.config.index_name = index_name
            self.save_config()

    def _get_index_names(self):
        return self._client.indices.get_alias().keys()

    def _get_index_ids(self, batch_size=1000):
        sample_ids = []
        label_ids = []
        for batch in range(0, self.total_index_size, batch_size):
            response = self._client.search(
                index=self.config.index_name,
                body={
                    "fields": ["sample_id"],
                    "from": batch,
                    "query": {
                        "bool": {
                            "must": [
                                {"exists": {"field": "vector"}},
                                {"exists": {"field": "sample_id"}},
                            ]
                        }
                    },
                },
                source=False,
                size=batch_size,
            )
            for doc in response["hits"]["hits"]:
                sample_id = doc["fields"]["sample_id"][0]
                sample_or_label_id = doc["_id"]
                sample_ids.append(sample_id)
                label_ids.append(sample_or_label_id)

        return sample_ids, label_ids

    def _get_dimension(self):
        if self.total_index_size == 0:
            return None

        if self.config.patches_field is not None:
            embeddings, _, _ = self.get_embeddings(
                label_ids=self._label_ids[:1]
            )
        else:
            embeddings, _, _ = self.get_embeddings(
                sample_ids=self._sample_ids[:1]
            )

        return embeddings.shape[1]

    def _get_metric(self):
        if self._metric is None:
            try:
                # We must ask ES rather than using `self.config.metric` because
                # we may be working with a preexisting index
                self._metric = self._client.indices.get_mapping(
                    index=self.config.index_name
                )[self.config.index_name]["mappings"]["properties"]["vector"][
                    "similarity"
                ]
            except:
                logger.warning(
                    "Failed to infer similarity metric from index '%s'",
                    self.config.index_name,
                )

        return self._metric

    def _index_exists(self):
        if self.config.index_name is None:
            return False

        return self.config.index_name in self._get_index_names()

    def _create_index(self, dimension):
        metric = _SUPPORTED_METRICS[self.config.metric]
        mappings = {
            "properties": {
                "vector": {
                    "type": "dense_vector",
                    "dims": dimension,
                    "index": "true",
                    "similarity": metric,
                }
            }
        }
        self._client.indices.create(
            index=self.config.index_name, mappings=mappings
        )
        self._metric = metric

    def _get_existing_ids(self, ids):
        docs = [{"_index": self.config.index_name, "_id": i} for i in ids]
        resp = self._client.mget(docs=docs)
        return [d["_id"] for d in resp["docs"] if d["found"]]

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
        batch_size=500,
    ):
        if not self._index_exists():
            self._create_index(embeddings.shape[1])

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            existing_ids = self._get_existing_ids(ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            del_inds = [i for i, _id in enumerate(ids) if _id in existing_ids]
            embeddings = np.delete(embeddings, del_inds, axis=0)
            sample_ids = np.delete(sample_ids, del_inds)
            if label_ids is not None:
                label_ids = np.delete(label_ids, del_inds)

        if self._get_metric() == _SUPPORTED_METRICS["dotproduct"]:
            embeddings /= np.linalg.norm(embeddings, axis=1)[:, np.newaxis]

        embeddings = [e.tolist() for e in embeddings]
        sample_ids = list(sample_ids)
        if label_ids is not None:
            ids = list(label_ids)
        else:
            ids = list(sample_ids)

        for _embeddings, _ids, _sample_ids in zip(
            fou.iter_batches(embeddings, batch_size),
            fou.iter_batches(ids, batch_size),
            fou.iter_batches(sample_ids, batch_size),
        ):
            operations = []
            for _e, _id, _sid in zip(_embeddings, _ids, _sample_ids):
                operations.append(
                    {"index": {"_index": self.config.index_name, "_id": _id}}
                )
                operations.append({"sample_id": _sid, "vector": _e})

            self._client.bulk(
                index=self.config.index_name,
                operations=operations,
                refresh=True,
            )

        if reload:
            self.reload()

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if not allow_missing or warn_missing:
            existing_ids = self._get_existing_ids(ids)
            missing_ids = set(ids) - set(existing_ids)
            num_missing = len(missing_ids)

            if num_missing > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that are not present in the "
                        "index" % (num_missing, next(iter(missing_ids)))
                    )

                if warn_missing:
                    logger.warning(
                        "Ignoring %d IDs that are not present in the index",
                        num_missing,
                    )

                ids = existing_ids

        operations = [
            {"delete": {"_index": self.config.index_name, "_id": i}}
            for i in ids
        ]
        self._client.bulk(body=operations, refresh=True)

        if reload:
            self.reload()

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        if sample_ids is not None and self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_sample_ids(sample_ids)
        elif self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_label_ids(label_ids)
        else:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_sample_embeddings(sample_ids)

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(embeddings)
        sample_ids = np.array(sample_ids)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return embeddings, sample_ids, label_ids

    def _parse_embeddings_response(self, response, label_id=True):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []
        for r in response:
            if r.get("found", True):
                found_embeddings.append(r["_source"]["vector"])
                if label_id:
                    found_sample_ids.append(r["_source"]["sample_id"])
                    found_label_ids.append(r["_id"])
                else:
                    found_sample_ids.append(r["_id"])

        return found_embeddings, found_sample_ids, found_label_ids

    def _get_sample_embeddings(self, sample_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []

        if sample_ids is None:
            sample_ids, label_ids = self._get_index_ids(batch_size=batch_size)

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            response = self._client.mget(
                index=self.config.index_name, ids=batch_ids, source=True
            )

            (
                _found_embeddings,
                _found_sample_ids,
                _,
            ) = self._parse_embeddings_response(
                response["docs"], label_id=False
            )
            found_embeddings += _found_embeddings
            found_sample_ids += _found_sample_ids

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, None, missing_ids

    def _get_patch_embeddings_from_label_ids(self, label_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        if label_ids is None:
            sample_ids, label_ids = self._get_index_ids(batch_size=batch_size)

        for batch_ids in fou.iter_batches(label_ids, batch_size):
            response = self._client.mget(
                index=self.config.index_name, ids=batch_ids, source=True
            )

            (
                _found_embeddings,
                _found_sample_ids,
                _found_label_ids,
            ) = self._parse_embeddings_response(response["docs"])
            found_embeddings += _found_embeddings
            found_sample_ids += _found_sample_ids
            found_label_ids += _found_label_ids

        missing_ids = list(set(label_ids) - set(found_label_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _get_patch_embeddings_from_sample_ids(
        self, sample_ids, batch_size=100
    ):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        if sample_ids is None:
            sample_ids, label_ids = self._get_index_ids(batch_size=batch_size)

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            response = self._client.search(
                index=self.config.index_name,
                body={"query": {"terms": {"sample_id": sample_ids}}},
            )

            (
                _found_embeddings,
                _found_sample_ids,
                _found_label_ids,
            ) = self._parse_embeddings_response(response["hits"]["hits"])
            found_embeddings += _found_embeddings
            found_sample_ids += _found_sample_ids
            found_label_ids += _found_label_ids

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def cleanup(self):
        self._client.indices.delete(
            index=self.config.index_name, ignore_unavailable=True
        )

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        if query is None:
            raise ValueError(
                "Elasticsearch does not support full index neighbors"
            )

        if reverse is True:
            raise ValueError(
                "Elasticsearch does not support least similarity queries"
            )

        if aggregation not in (None, "mean"):
            raise ValueError(
                f"Elasticsearch does not support {aggregation} aggregation"
            )

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        if self.has_view:
            if self.config.patches_field is not None:
                index_ids = self.current_label_ids
            else:
                index_ids = self.current_sample_ids

            _filter = {"terms": {"_id": list(index_ids)}}
        else:
            _filter = None

        ids = []
        dists = []
        for q in query:
            if self._get_metric() == _SUPPORTED_METRICS["dotproduct"]:
                q /= np.linalg.norm(q)

            knn = {
                "field": "vector",
                "query_vector": q.tolist(),
                "k": k,
                "num_candidates": 10 * k,
            }
            if _filter:
                knn["filter"] = _filter

            response = self._client.search(
                index=self.config.index_name,
                knn=knn,
                size=k,
            )
            ids.append([r["_id"] for r in response["hits"]["hits"]])
            if return_dists:
                dists.append([r["_score"] for r in response["hits"]["hits"]])

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        response = self._client.mget(
            index=self.config.index_name, ids=query_ids, source=True
        )
        query = np.array(
            [r["_source"]["vector"] for r in response["docs"] if r["found"]]
        )

        if single_query:
            query = query[0, :]

        return query

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)



"""
Similarity interface.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from copy import deepcopy
import inspect
import logging

from bson import ObjectId
import numpy as np

import eta.core.utils as etau

import fiftyone.brain as fb
import fiftyone.core.brain as fob
import fiftyone.core.context as foc
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.patches as fop
import fiftyone.core.stages as fos
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov
import fiftyone.zoo as foz
from fiftyone import ViewField as F

fbu = fou.lazy_import("fiftyone.brain.internal.core.utils")


logger = logging.getLogger(__name__)

_ALLOWED_ROI_FIELD_TYPES = (
    fol.Detection,
    fol.Detections,
    fol.Polyline,
    fol.Polylines,
)

_DEFAULT_MODEL = "mobilenet-v2-imagenet-torch"
_DEFAULT_BATCH_SIZE = None


def compute_similarity(
    samples,
    patches_field,
    roi_field,
    embeddings,
    brain_key,
    model,
    model_kwargs,
    force_square,
    alpha,
    batch_size,
    num_workers,
    skip_failures,
    progress,
    backend,
    **kwargs,
):
    """See ``fiftyone/brain/__init__.py``."""

    fov.validate_collection(samples)

    if roi_field is not None:
        fov.validate_collection_label_fields(
            samples, roi_field, _ALLOWED_ROI_FIELD_TYPES
        )

    # Allow for `embeddings_field=XXX` and `embeddings=False` together
    embeddings_field = kwargs.pop("embeddings_field", None)
    if embeddings_field is not None or etau.is_str(embeddings):
        if embeddings_field is None:
            embeddings_field = embeddings
            embeddings = None

        embeddings_field, embeddings_exist = fbu.parse_data_field(
            samples,
            embeddings_field,
            patches_field=patches_field or roi_field,
            data_type="embeddings",
        )
    else:
        embeddings_field = None
        embeddings_exist = None

    if model is None and embeddings is None and not embeddings_exist:
        model = _DEFAULT_MODEL
        if batch_size is None:
            batch_size = _DEFAULT_BATCH_SIZE

    if etau.is_str(model):
        _model_kwargs = model_kwargs or {}
        _model = foz.load_zoo_model(model, **_model_kwargs)
        try:
            supports_prompts = _model.can_embed_prompts
        except:
            supports_prompts = None
    else:
        _model = model
        supports_prompts = None

    config = _parse_config(
        backend,
        embeddings_field=embeddings_field,
        patches_field=patches_field,
        roi_field=roi_field,
        model=model,
        model_kwargs=model_kwargs,
        supports_prompts=supports_prompts,
        **kwargs,
    )
    brain_method = config.build()
    brain_method.ensure_requirements()

    # Similarity indexes can be modified after creation, so we always register
    # the index on the full dataset so that queries will always be performed
    # against the full index by default
    dataset = samples._root_dataset

    if brain_key is not None:
        # Don't allow overwriting an existing run with same key, since we
        # need the existing run in order to perform workflows like
        # automatically cleaning up the backend's index
        brain_method.register_run(dataset, brain_key, overwrite=False)

    results = brain_method.initialize(dataset, brain_key)

    get_embeddings = embeddings is not False
    if not results.is_external and results.total_index_size > 0:
        # No need to load embeddings because the index already has them
        get_embeddings = False

    if get_embeddings:
        # Don't immediatly store embeddings in DB; let `add_to_index()` do it
        if not embeddings_exist:
            embeddings_field = None

        if roi_field is not None:
            handle_missing = "image"
            agg_fcn = lambda e: np.mean(e, axis=0)
        else:
            handle_missing = "skip"
            agg_fcn = None

        embeddings, sample_ids, label_ids = fbu.get_embeddings(
            samples,
            model=_model,
            patches_field=patches_field or roi_field,
            embeddings=embeddings,
            embeddings_field=embeddings_field,
            force_square=force_square,
            alpha=alpha,
            handle_missing=handle_missing,
            agg_fcn=agg_fcn,
            batch_size=batch_size,
            num_workers=num_workers,
            skip_failures=skip_failures,
            progress=progress,
        )
    else:
        embeddings = None

    if embeddings is not None:
        results.add_to_index(embeddings, sample_ids, label_ids=label_ids)

    brain_method.save_run_results(dataset, brain_key, results)

    return results


def _parse_config(name, **kwargs):
    if name is None:
        name = fb.brain_config.default_similarity_backend

    if inspect.isclass(name):
        return name(**kwargs)

    backends = fb.brain_config.similarity_backends

    if name not in backends:
        raise ValueError(
            "Unsupported backend '%s'. The available backends are %s"
            % (name, sorted(backends.keys()))
        )

    params = deepcopy(backends[name])

    config_cls = kwargs.pop("config_cls", None)

    if config_cls is None:
        config_cls = params.pop("config_cls", None)

    if config_cls is None:
        raise ValueError("Similarity backend '%s' has no `config_cls`" % name)

    if etau.is_str(config_cls):
        config_cls = etau.get_class(config_cls)

    params.update(**kwargs)
    return config_cls(**params)


class SimilarityConfig(fob.BrainMethodConfig):
    """Similarity configuration.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or name of the
            zoo model that was used to compute embeddings, if known
        model_kwargs (None): a dictionary of optional keyword arguments to pass
            to the model's ``Config`` when a model name is provided
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        roi_field (None): the sample field defining a region of interest within
            each image to use to compute embeddings, if any
        supports_prompts (False): whether this run supports prompt queries
    """

    def __init__(
        self,
        embeddings_field=None,
        model=None,
        model_kwargs=None,
        patches_field=None,
        roi_field=None,
        supports_prompts=None,
        **kwargs,
    ):
        if model is not None and not etau.is_str(model):
            model = None

        self.embeddings_field = embeddings_field
        self.model = model
        self.model_kwargs = model_kwargs
        self.patches_field = patches_field
        self.roi_field = roi_field
        self.supports_prompts = supports_prompts
        super().__init__(**kwargs)

    @property
    def type(self):
        return "similarity"

    @property
    def method(self):
        """The name of the similarity backend."""
        raise NotImplementedError("subclass must implement method")

    @property
    def max_k(self):
        """A maximum k value for nearest neighbor queries, or None if there is
        no limit.
        """
        raise NotImplementedError("subclass must implement max_k")

    @property
    def supports_least_similarity(self):
        """Whether this backend supports least similarity queries."""
        raise NotImplementedError(
            "subclass must implement supports_least_similarity"
        )

    @property
    def supported_aggregations(self):
        """A tuple of supported values for the ``aggregation`` parameter of the
        backend's
        :meth:`sort_by_similarity() <SimilarityIndex.sort_by_similarity>` and
        :meth:`_kneighbors() <SimilarityIndex._kneighbors>` methods.
        """
        raise NotImplementedError(
            "subclass must implement supported_aggregations"
        )

    def load_credentials(self, **kwargs):
        self._load_parameters(**kwargs)

    def _load_parameters(self, **kwargs):
        name = self.method
        parameters = fb.brain_config.similarity_backends.get(name, {})

        for name, value in kwargs.items():
            if value is None:
                value = parameters.get(name, None)

            if value is not None:
                setattr(self, name, value)


class Similarity(fob.BrainMethod):
    """Base class for similarity factories.

    Args:
        config: a :class:`SimilarityConfig`
    """

    def initialize(self, samples, brain_key):
        """Initializes a similarity index.

        Args:
            samples: a :class:`fiftyone.core.collections.SampleColllection`
            brain_key: the brain key

        Returns:
            a :class:`SimilarityIndex`
        """
        raise NotImplementedError("subclass must implement initialize()")

    def get_fields(self, samples, brain_key):
        fields = []
        if self.config.patches_field is not None:
            fields.append(self.config.patches_field)

        if self.config.embeddings_field is not None:
            fields.append(self.config.embeddings_field)

        return fields


class SimilarityIndex(fob.BrainResults):
    """Base class for similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`SimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`Similarity` backend
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)

        self._model = None
        self._curr_view = None
        self._curr_sample_ids = None
        self._curr_label_ids = None
        self._curr_keep_inds = None
        self._curr_missing_size = None
        self._last_view = None
        self._last_views = []

        self.use_view(samples)

    def __enter__(self):
        self._last_views.append(self._last_view)
        return self

    def __exit__(self, *args):
        try:
            last_view = self._last_views.pop()
        except:
            last_view = self._samples

        self.use_view(last_view)

    @property
    def config(self):
        """The :class:`SimilarityConfig` for these results."""
        return self._config

    @property
    def is_external(self):
        """Whether this similarity index manages its own embeddings (True) or
        loads them directly from the ``embeddings_field`` of the dataset
        (False).
        """
        return True  # assume external unless explicitly overridden

    @property
    def sample_ids(self):
        """The sample IDs of the full index, or ``None`` if not supported."""
        return None

    @property
    def label_ids(self):
        """The label IDs of the full index, or ``None`` if not applicable or
        not supported.
        """
        return None

    @property
    def total_index_size(self):
        """The total number of data points in the index.

        If :meth:`use_view` has been called to restrict the index, this value
        may be larger than the current :meth:`index_size`.
        """
        raise NotImplementedError("subclass must implement total_index_size")

    @property
    def has_view(self):
        """Whether the index is currently restricted to a view.

        Use :meth:`use_view` to restrict the index to a view, and use
        :meth:`clear_view` to reset to the full index.
        """
        return self._curr_view.view() != self._samples.view()

    @property
    def view(self):
        """The :class:`fiftyone.core.collections.SampleCollection` against
        which results are currently being generated.

        If :meth:`use_view` has been called, this view may be different than
        the collection on which the full index was generated.
        """
        return self._curr_view

    @property
    def current_sample_ids(self):
        """The sample IDs of the currently active data points in the index.

        If :meth:`use_view` has been called, this may be a subset of the full
        index.
        """
        return self._curr_sample_ids

    @property
    def current_label_ids(self):
        """The label IDs of the currently active data points in the index, or
        ``None`` if not applicable.

        If :meth:`use_view` has been called, this may be a subset of the full
        index.
        """
        return self._curr_label_ids

    @property
    def _current_inds(self):
        """The indices of :meth:`current_sample_ids` in :meth:`sample_ids`, or
        ``None`` if not supported or if the full index is currently being used.
        """
        return self._curr_keep_inds

    @property
    def index_size(self):
        """The number of active data points in the index.

        If :meth:`use_view` has been called to restrict the index, this
        property will reflect the size of the active index.
        """
        return len(self._curr_sample_ids)

    @property
    def missing_size(self):
        """The total number of data points in :meth:`view` that are missing
        from this index, or ``None`` if unknown.

        This property is only applicable when :meth:`use_view` has been called,
        and it will be ``None`` if no data points are missing or when the
        backend does not support it.
        """
        return self._curr_missing_size

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
    ):
        """Adds the given embeddings to the index.

        Args:
            embeddings: a ``num_embeddings x num_dims`` array of embeddings
            sample_ids: a ``num_embeddings`` array of sample IDs
            label_ids (None): a ``num_embeddings`` array of label IDs, if
                applicable
            overwrite (True): whether to replace (True) or ignore (False)
                existing embeddings with the same sample/label IDs
            allow_existing (True): whether to ignore (True) or raise an error
                (False) when ``overwrite`` is False and a provided ID already
                exists in the
            warn_existing (False): whether to log a warning if an embedding is
                not added to the index because its ID already exists
            reload (True): whether to call :meth:`reload` to refresh the
                current view after the update
        """
        raise NotImplementedError("subclass must implement add_to_index()")

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        """Removes the specified embeddings from the index.

        Args:
            sample_ids (None): an array of sample IDs
            label_ids (None): an array of label IDs, if applicable
            allow_missing (True): whether to allow the index to not contain IDs
                that you provide (True) or whether to raise an error in this
                case (False)
            warn_missing (False): whether to log a warning if the index does
                not contain IDs that you provide
            reload (True): whether to call :meth:`reload` to refresh the
                current view after the update
        """
        raise NotImplementedError(
            "subclass must implement remove_from_index()"
        )

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        """Retrieves the embeddings for the given IDs from the index.

        If no IDs are provided, the entire index is returned.

        Args:
            sample_ids (None): a sample ID or list of sample IDs for which to
                retrieve embeddings
            label_ids (None): a label ID or list of label IDs for which to
                retrieve embeddings
            allow_missing (True): whether to allow the index to not contain IDs
                that you provide (True) or whether to raise an error in this
                case (False)
            warn_missing (False): whether to log a warning if the index does
                not contain IDs that you provide

        Returns:
            a tuple of:

            -   a ``num_embeddings x num_dims`` array of embeddings
            -   a ``num_embeddings`` array of sample IDs
            -   a ``num_embeddings`` array of label IDs, if applicable, or else
                ``None``
        """
        raise NotImplementedError("subclass must implement get_embeddings()")

    def use_view(self, samples, allow_missing=True, warn_missing=False):
        """Restricts the index to the provided view.

        Subsequent calls to methods on this instance will only contain results
        from the specified view rather than the full index.

        Use :meth:`clear_view` to reset to the full index. Or, equivalently,
        use the context manager interface as demonstrated below to
        automatically reset the view when the context exits.

        Example usage::

            import fiftyone as fo
            import fiftyone.brain as fob
            import fiftyone.zoo as foz

            dataset = foz.load_zoo_dataset("quickstart")

            results = fob.compute_similarity(dataset)
            print(results.index_size)  # 200

            view = dataset.take(50)

            with results.use_view(view):
                print(results.index_size)  # 50

                results.find_unique(10)
                print(results.unique_ids)

                plot = results.visualize_unique()
                plot.show()

        Args:
            samples: a :class:`fiftyone.core.collections.SampleCollection`
            allow_missing (True): whether to allow the provided collection to
                contain data points that this index does not contain (True) or
                whether to raise an error in this case (False)
            warn_missing (False): whether to log a warning if the provided
                collection contains data points that this index does not
                contain

        Returns:
            self
        """
        sample_ids, label_ids, keep_inds, good_inds = fbu.filter_ids(
            samples,
            self.sample_ids,
            self.label_ids,
            patches_field=self.config.patches_field,
            allow_missing=allow_missing,
            warn_missing=warn_missing,
        )

        if good_inds is not None:
            missing_size = good_inds.size - np.count_nonzero(good_inds)
        else:
            missing_size = None

        self._last_view = self._curr_view
        self._curr_view = samples
        self._curr_sample_ids = sample_ids
        self._curr_label_ids = label_ids
        self._curr_keep_inds = keep_inds
        self._curr_missing_size = missing_size

        return self

    def clear_view(self):
        """Clears the view set by :meth:`use_view`, if any.

        Subsequent operations will be performed on the full index.
        """
        self.use_view(self._samples)

    def reload(self):
        """Reloads the index for the current view.

        Subclasses may override this method, but by default this method simply
        passes the current :meth:`view` back into :meth:`use_view`, which
        updates the index's current ID set based on any changes to the view
        since the index was last loaded.
        """
        self.use_view(self._curr_view)

    def cleanup(self):
        """Deletes the similarity index from the backend."""
        raise NotImplementedError("subclass must implement cleanup()")

    def values(self, path_or_expr):
        """Extracts a flat list of values from the given field or expression
        corresponding to the current :meth:`view`.

        This method always returns values in the same order as
        :meth:`current_sample_ids` and :meth:`current_label_ids`.

        Args:
            path_or_expr: the values to extract, which can be:

                -   the name of a sample field or ``embedded.field.name`` from
                    which to extract numeric or string values
                -   a :class:`fiftyone.core.expressions.ViewExpression`
                    defining numeric or string values to compute via
                    :meth:`fiftyone.core.collections.SampleCollection.values`

        Returns:
            a list of values
        """
        samples = self.view
        patches_field = self.config.patches_field

        if patches_field is not None:
            ids = self.current_label_ids
        else:
            ids = self.current_sample_ids

        return fbu.get_values(
            samples, path_or_expr, ids, patches_field=patches_field
        )

    def sort_by_similarity(
        self,
        query,
        k=None,
        reverse=False,
        aggregation="mean",
        dist_field=None,
        _mongo=False,
    ):
        """Returns a view that sorts the samples/labels in :meth:`view` by
        similarity to the specified query.

        When querying by IDs, the query can be any ID(s) in the full index of
        this instance, even if the current :meth:`view` contains a subset of
        the full index.

        Args:
            query: the query, which can be any of the following:

                -   an ID or iterable of IDs
                -   a ``num_dims`` vector or ``num_queries x num_dims`` array
                    of vectors
                -   a prompt or iterable of prompts (if supported by the index)

            k (None): the number of matches to return. Some backends may
                support ``None``, in which case all samples will be sorted
            reverse (False): whether to sort by least similarity (True) or
                greatest similarity (False). Some backends may not support
                least similarity
            aggregation ("mean"): the aggregation method to use when multiple
                queries are provided. The default is ``"mean"``, which means
                that the query vectors are averaged prior to searching. Some
                backends may support additional options
            dist_field (None): the name of a float field in which to store the
                distance of each example to the specified query. The field is
                created if necessary

        Returns:
            a :class:`fiftyone.core.view.DatasetView`
        """
        samples = self.view
        patches_field = self.config.patches_field

        selecting_samples = patches_field is None or isinstance(
            samples, fop.PatchesView
        )

        kwargs = dict(
            query=self._parse_query(query),
            k=k,
            reverse=reverse,
            aggregation=aggregation,
            return_dists=dist_field is not None,
        )

        if dist_field is not None:
            ids, dists = self._kneighbors(**kwargs)
        else:
            ids = self._kneighbors(**kwargs)

        if not selecting_samples:
            label_ids = ids

            _ids = set(ids)
            bools = np.array([_id in _ids for _id in self.current_label_ids])
            sample_ids = self.current_sample_ids[bools]

        # Store query distances
        if dist_field is not None:
            if selecting_samples:
                values = dict(zip(ids, dists))
                samples.set_values(dist_field, values, key_field="id")
            else:
                label_type, path = samples._get_label_field_path(
                    patches_field, dist_field
                )
                if issubclass(label_type, fol._LABEL_LIST_FIELDS):
                    samples._set_list_values_by_id(
                        path,
                        sample_ids,
                        label_ids,
                        dists,
                        path.rsplit(".", 1)[0],
                    )
                else:
                    values = dict(zip(sample_ids, dists))
                    samples.set_values(path, values, key_field="id")

        # Construct sorted view
        stages = []

        if selecting_samples:
            stage = fos.Select(ids, ordered=True)
            stages.append(stage)
        else:
            # Sorting by object similarity but this is not a patches view, so
            # arrange the samples in order of their first occuring label
            result_sample_ids = _unique_no_sort(sample_ids)
            stage = fos.Select(result_sample_ids, ordered=True)
            stages.append(stage)

            if k is not None:
                _ids = [ObjectId(_id) for _id in ids]
                stage = fos.FilterLabels(patches_field, F("_id").is_in(_ids))
                stages.append(stage)

        if _mongo:
            pipeline = []
            for stage in stages:
                stage.validate(samples)
                pipeline.extend(stage.to_mongo(samples))

            return pipeline

        view = samples
        for stage in stages:
            view = view.add_stage(stage)

        return view

    def _parse_query(self, query):
        if query is None:
            raise ValueError("At least one query must be provided")

        if isinstance(query, np.ndarray):
            # Query by vector(s)
            if query.size == 0:
                raise ValueError("At least one query vector must be provided")

            return query

        if etau.is_str(query):
            query = [query]
        else:
            query = list(query)

        if not query:
            raise ValueError("At least one query must be provided")

        if etau.is_numeric(query[0]):
            return np.asarray(query)

        try:
            ObjectId(query[0])
            is_prompts = False
        except:
            is_prompts = True

        if is_prompts:
            if not self.config.supports_prompts:
                raise ValueError(
                    "Invalid query '%s'; this model does not support prompts"
                    % query[0]
                )

            model = self.get_model()
            with model:
                return model.embed_prompts(query)

        return query

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        """Returns the k-nearest neighbors for the given query.

        This method should only return results from the current :meth:`view`.

        Args:
            query (None): the query, which can be any of the following:

                -   an ID or list of IDs for which to return neighbors
                -   an embedding or ``num_queries x num_dim`` array of
                    embeddings for which to return neighbors
                -   Some backends may also support ``None``, in which case the
                    neighbors for all points in the current :meth:`view are
                    returned

            k (None): the number of neighbors to return. Some backends may
                enforce upper bounds on this parameter
            reverse (False): whether to sort by least similarity (True) or
                greatest similarity (False). Some backends may not support
                least similarity
            aggregation (None): an optional aggregation method to use when
                multiple queries are provided. All backends must support
                ``"mean"``, which averages query vectors prior to searching.
                Backends may support additional options as well
            return_dists (False): whether to return query-neighbor distances

        Returns:
            the query result, in one of the following formats:

                -   an ``(ids, dists)`` tuple, when ``return_dists`` is True
                -   ``ids``, when ``return_dists`` is False

            In the above, ``ids`` contains the IDs of the nearest neighbors, in
            one of the following formats:

                -   a list of nearest neighbor IDs, when a single query ID or
                    vector is provided, **or** when an ``aggregation`` is
                    provided
                -   a list of lists of nearest neighbor IDs, when multiple
                    query IDs/vectors and no ``aggregation`` is provided
                -   a list of arrays of the **integer indexes** (not IDs) of
                    nearest neighbor points for every vector in the index, when
                    no query is provided

            and ``dists`` contains the corresponding query-neighbor distances
            for each result in ``ids``
        """
        raise NotImplementedError("subclass must implement _kneighbors()")

    def get_model(self):
        """Returns the stored model for this index.

        Returns:
            a :class:`fiftyone.core.models.Model`
        """
        if self._model is None:
            model = self.config.model
            if model is None:
                raise ValueError("These results don't have a stored model")

            if etau.is_str(model):
                model_kwargs = self.config.model_kwargs or {}
                model = foz.load_zoo_model(model, **model_kwargs)

            self._model = model

        return self._model

    def compute_embeddings(
        self,
        samples,
        model=None,
        batch_size=None,
        num_workers=None,
        skip_failures=True,
        skip_existing=False,
        warn_existing=False,
        force_square=False,
        alpha=None,
        progress=None,
    ):
        """Computes embeddings for the given samples using this backend's
        model.

        Args:
            samples: a :class:`fiftyone.core.collections.SampleCollection`
            model (None): a :class:`fiftyone.core.models.Model` to apply. If
                not provided, these results must have been created with a
                stored model, which will be used by default
            batch_size (None): an optional batch size to use when computing
                embeddings. Only applicable when a ``model`` is provided
            num_workers (None): the number of workers to use when loading
                images. Only applicable when a Torch-based model is being used
                to compute embeddings
            skip_failures (True): whether to gracefully continue without
                raising an error if embeddings cannot be generated for a sample
            skip_existing (False): whether to skip generating embeddings for
                sample/label IDs that are already in the index
            warn_existing (False): whether to log a warning if any IDs already
                exist in the index
            force_square (False): whether to minimally manipulate the patch
                bounding boxes into squares prior to extraction. Only
                applicable when a ``model`` and ``patches_field`` are specified
            alpha (None): an optional expansion/contraction to apply to the
                patches before extracting them, in ``[-1, inf)``. If provided,
                the length and width of the box are expanded (or contracted,
                when ``alpha < 0``) by ``(100 * alpha)%``. For example, set
                ``alpha = 0.1`` to expand the boxes by 10%, and set
                ``alpha = -0.1`` to contract the boxes by 10%. Only applicable
                when a ``model`` and ``patches_field`` are specified
            progress (None): whether to render a progress bar (True/False), use
                the default value ``fiftyone.config.show_progress_bars``
                (None), or a progress callback function to invoke instead

        Returns:
            a tuple of:

            -   a ``num_embeddings x num_dims`` array of embeddings
            -   a ``num_embeddings`` array of sample IDs
            -   a ``num_embeddings`` array of label IDs, if applicable, or else
                ``None``
        """
        if model is None:
            model = self.get_model()

        if skip_existing:
            if self.config.patches_field is not None:
                index_ids = self.label_ids
            else:
                index_ids = self.sample_ids

            if index_ids is not None:
                samples = fbu.skip_ids(
                    samples,
                    index_ids,
                    patches_field=self.config.patches_field,
                    warn_existing=warn_existing,
                )
            else:
                logger.warning(
                    "This index does not support skipping existing IDs"
                )

        if self.config.roi_field is not None:
            patches_field = self.config.roi_field
            handle_missing = "image"
            agg_fcn = lambda e: np.mean(e, axis=0)
        else:
            patches_field = self.config.patches_field
            handle_missing = "skip"
            agg_fcn = None

        return fbu.get_embeddings(
            samples,
            model=model,
            patches_field=patches_field,
            force_square=force_square,
            alpha=alpha,
            handle_missing=handle_missing,
            agg_fcn=agg_fcn,
            batch_size=batch_size,
            num_workers=num_workers,
            skip_failures=skip_failures,
            progress=progress,
        )

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        """Builds a :class:`SimilarityIndex` from a JSON representation of it.

        Args:
            d: a JSON dict
            samples: the :class:`fiftyone.core.collections.SampleCollection`
                for the run
            config: the :class:`SimilarityConfig` for the run
            brain_key: the brain key

        Returns:
            a :class:`SimilarityIndex`
        """
        raise NotImplementedError("subclass must implement _from_dict()")


class DuplicatesMixin(object):
    """Mixin for :class:`SimilarityIndex` instances that support duplicate
    detection operations.

    Similarity backends can expose this mixin simply by implementing
    :meth:`_radius_neighbors`.
    """

    def __init__(self):
        self._thresh = None
        self._unique_ids = None
        self._duplicate_ids = None
        self._neighbors_map = None

    @property
    def thresh(self):
        """The threshold used by the last call to :meth:`find_duplicates` or
        :meth:`find_unique`.
        """
        return self._thresh

    @property
    def unique_ids(self):
        """A list of unique IDs from the last call to :meth:`find_duplicates`
        or :meth:`find_unique`.
        """
        return self._unique_ids

    @property
    def duplicate_ids(self):
        """A list of duplicate IDs from the last call to
        :meth:`find_duplicates` or :meth:`find_unique`.
        """
        return self._duplicate_ids

    @property
    def neighbors_map(self):
        """A dictionary mapping IDs to lists of ``(dup_id, dist)`` tuples from
        the last call to :meth:`find_duplicates`.
        """
        return self._neighbors_map

    def _radius_neighbors(self, query=None, thresh=None, return_dists=False):
        """Returns the neighbors within the given distance threshold for the
        given query.

        This method should only return results from the current :meth:`view`.

        Args:
            query (None): the query, which can be any of the following:

                -   an ID or list of IDs for which to return neighbors
                -   an embedding or ``num_queries x num_dim`` array of
                    embeddings for which to return neighbors
                -   ``None``, in which case the neighbors for all points in the
                    current :meth:`view are returned

            thresh (None): the distance threshold to use
            return_dists (False): whether to return query-neighbor distances

        Returns:
            the query result, in one of the following formats:

                -   an ``(ids, dists)`` tuple, when ``return_dists`` is True
                -   ``ids``, when ``return_dists`` is False

            In the above, ``ids`` contains the IDs of the nearest neighbors, in
            one of the following formats:

                -   a list of nearest neighbor IDs, when a single query ID or
                    vector is provided
                -   a list of lists of nearest neighbor IDs, when multiple
                    query IDs/vectors is provided
                -   a list of arrays of the **integer indexes** (not IDs) of
                    nearest neighbor points for every vector in the index, when
                    no query is provided

            and ``dists`` contains the corresponding query-neighbor distances
            for each result in ``ids``
        """
        raise NotImplementedError(
            "subclass must implement _radius_neighbors()"
        )

    def find_duplicates(self, thresh=None, fraction=None):
        """Queries the index to find near-duplicate examples based on the
        provided parameters.

        Calling this method populates the :meth:`unique_ids`,
        :meth:`duplicate_ids`, :attr:`neighbors_map`, and :attr:`thresh`
        properties of this object with the results of the query.

        Use :meth:`duplicates_view` and :meth:`visualize_duplicates` to analyze
        the results generated by this method.

        Args:
            thresh (None): a distance threshold to use to determine duplicates.
                If specified, the non-duplicate set will be the (approximately)
                largest set such that all pairwise distances between
                non-duplicate examples are greater than this threshold
            fraction (None): a desired fraction of images/patches to tag as
                duplicates, in ``[0, 1]``. In this case ``thresh`` is
                automatically tuned to achieve the desired fraction of
                duplicates
        """
        if self.config.patches_field is not None:
            logger.info("Computing duplicate patches...")
            ids = self.current_label_ids
        else:
            logger.info("Computing duplicate samples...")
            ids = self.current_sample_ids

        # Detect duplicates
        if fraction is not None:
            num_keep = int(round(min(max(0, 1.0 - fraction), 1) * len(ids)))
            unique_ids, thresh = self._remove_duplicates_count(
                num_keep, ids, init_thresh=thresh
            )
        else:
            unique_ids = self._remove_duplicates_thresh(thresh, ids)

        _unique_ids = set(unique_ids)
        duplicate_ids = [_id for _id in ids if _id not in _unique_ids]

        # Locate nearest non-duplicate for each duplicate
        if unique_ids and duplicate_ids:
            if self.config.patches_field is not None:
                unique_view = self._samples.select_labels(
                    ids=unique_ids, fields=self.config.patches_field
                )
            else:
                unique_view = self._samples.select(unique_ids)

            with self.use_view(unique_view):
                nearest_ids, dists = self._kneighbors(
                    query=duplicate_ids, k=1, return_dists=True
                )

            neighbors_map = defaultdict(list)
            for dup_id, _ids, _dists in zip(duplicate_ids, nearest_ids, dists):
                neighbors_map[_ids[0]].append((dup_id, _dists[0]))

            neighbors_map = {
                k: sorted(v, key=lambda t: t[1])
                for k, v in neighbors_map.items()
            }
        else:
            neighbors_map = {}

        logger.info("Duplicates computation complete")

        self._thresh = thresh
        self._unique_ids = unique_ids
        self._duplicate_ids = duplicate_ids
        self._neighbors_map = neighbors_map

    def find_unique(self, count):
        """Queries the index to select a subset of examples of the specified
        size that are maximally unique with respect to each other.

        Calling this method populates the :meth:`unique_ids`,
        :meth:`duplicate_ids`, and :attr:`thresh` properties of this object
        with the results of the query.

        Use :meth:`unique_view` and :meth:`visualize_unique` to analyze the
        results generated by this method.

        Args:
            count: the desired number of unique examples
        """
        if self.config.patches_field is not None:
            logger.info("Computing unique patches...")
            ids = self.current_label_ids
        else:
            logger.info("Computing unique samples...")
            ids = self.current_sample_ids

        unique_ids, thresh = self._remove_duplicates_count(count, ids)

        _unique_ids = set(unique_ids)
        duplicate_ids = [_id for _id in ids if _id not in _unique_ids]

        logger.info("Uniqueness computation complete")

        self._thresh = thresh
        self._unique_ids = unique_ids
        self._duplicate_ids = duplicate_ids
        self._neighbors_map = None

    def _remove_duplicates_count(self, num_keep, ids, init_thresh=None):
        if init_thresh is not None:
            thresh = init_thresh
        else:
            thresh = 1

        if num_keep <= 0:
            logger.info(
                "threshold: -, kept: %d, target: %d", num_keep, num_keep
            )
            return set(), None

        if num_keep >= len(ids):
            logger.info(
                "threshold: -, kept: %d, target: %d", num_keep, num_keep
            )
            return set(ids), None

        thresh_lims = [0, None]
        num_target = num_keep
        num_keep = -1

        while True:
            keep_ids = self._remove_duplicates_thresh(thresh, ids)
            num_keep_last = num_keep
            num_keep = len(keep_ids)

            logger.info(
                "threshold: %f, kept: %d, target: %d",
                thresh,
                num_keep,
                num_target,
            )

            if num_keep == num_target or (
                num_keep == num_keep_last
                and thresh_lims[1] is not None
                and thresh_lims[1] - thresh_lims[0] < 1e-6
            ):
                break

            if num_keep < num_target:
                # Need to decrease threshold
                thresh_lims[1] = thresh
                thresh = 0.5 * (thresh_lims[0] + thresh)
            else:
                # Need to increase threshold
                thresh_lims[0] = thresh
                if thresh_lims[1] is not None:
                    thresh = 0.5 * (thresh + thresh_lims[1])
                else:
                    thresh *= 2

        return keep_ids, thresh

    def _remove_duplicates_thresh(self, thresh, ids):
        nearest_inds = self._radius_neighbors(thresh=thresh)

        n = len(ids)
        keep = set(range(n))
        for ind in range(n):
            if ind in keep:
                keep -= {i for i in nearest_inds[ind] if i > ind}

        return [ids[i] for i in keep]

    def plot_distances(self, bins=100, log=False, backend="plotly", **kwargs):
        """Plots a histogram of the distance between each example and its
        nearest neighbor.

        If `:meth:`find_duplicates` or :meth:`find_unique` has been executed,
        the threshold used is also indicated on the plot.

        Args:
            bins (100): the number of bins to use
            log (False): whether to use a log scale y-axis
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method

        Returns:
            one of the following:

            -   a :class:`fiftyone.core.plots.plotly.PlotlyNotebookPlot`, if
                you are working in a notebook context and the plotly backend is
                used
            -   a plotly or matplotlib figure, otherwise
        """
        metric = self.config.metric
        thresh = self.thresh

        _, dists = self._kneighbors(k=1, return_dists=True)
        dists = np.array([d[0] for d in dists])

        if backend == "matplotlib":
            return _plot_distances_mpl(
                dists, metric, thresh, bins, log, **kwargs
            )

        return _plot_distances_plotly(
            dists, metric, thresh, bins, log, **kwargs
        )

    def duplicates_view(
        self,
        type_field=None,
        id_field=None,
        dist_field=None,
        sort_by="distance",
        reverse=False,
    ):
        """Returns a view that contains only the duplicate examples and their
        corresponding nearest non-duplicate examples generated by the last call
        to :meth:`find_duplicates`.

        If you are analyzing patches, the returned view will be a
        :class:`fiftyone.core.patches.PatchesView`.

        The examples are organized so that each non-duplicate is immediately
        followed by all duplicate(s) that are nearest to it.

        Args:
            type_field (None): the name of a string field in which to store
                ``"nearest"`` and ``"duplicate"`` labels. The field is created
                if necessary
            id_field (None): the name of a string field in which to store the
                ID of the nearest non-duplicate for each example in the view.
                The field is created if necessary
            dist_field (None): the name of a float field in which to store the
                distance of each example to its nearest non-duplicate example.
                The field is created if necessary
            sort_by ("distance"): specifies how to sort the groups of duplicate
                examples. The supported values are:

                -   ``"distance"``: sort the groups by the distance between the
                    non-duplicate and its (nearest, if multiple) duplicate
                -   ``"count"``: sort the groups by the number of duplicate
                    examples

            reverse (False): whether to sort in descending order

        Returns:
            a :class:`fiftyone.core.view.DatasetView`
        """
        if self.neighbors_map is None:
            raise ValueError(
                "You must first call `find_duplicates()` to generate results"
            )

        samples = self.view
        patches_field = self.config.patches_field
        neighbors_map = self.neighbors_map

        if patches_field is not None and not isinstance(
            samples, fop.PatchesView
        ):
            samples = samples.to_patches(patches_field)

        if sort_by == "distance":
            key = lambda kv: min(e[1] for e in kv[1])
        elif sort_by == "count":
            key = lambda kv: len(kv[1])
        else:
            raise ValueError(
                "Invalid sort_by='%s'; supported values are %s"
                % (sort_by, ("distance", "count"))
            )

        existing_ids = set(samples.values("id"))
        neighbors = [
            (k, v) for k, v in neighbors_map.items() if k in existing_ids
        ]

        ids = []
        types = {}
        nearest_ids = {}
        dists = {}
        for _id, duplicates in sorted(neighbors, key=key, reverse=reverse):
            ids.append(_id)
            types[_id] = "nearest"
            nearest_ids[_id] = _id
            dists[_id] = 0.0

            for dup_id, dist in duplicates:
                ids.append(dup_id)
                types[dup_id] = "duplicate"
                nearest_ids[dup_id] = _id
                dists[dup_id] = dist

        if type_field is not None:
            samples.set_values(type_field, types, key_field="id")

        if id_field is not None:
            samples.set_values(id_field, nearest_ids, key_field="id")

        if dist_field is not None:
            samples.set_values(dist_field, dists, key_field="id")

        return samples.select(ids, ordered=True)

    def unique_view(self):
        """Returns a view that contains only the unique examples generated by
        the last call to :meth:`find_duplicates` or :meth:`find_unique`.

        If you are analyzing patches, the returned view will be a
        :class:`fiftyone.core.patches.PatchesView`.

        Returns:
            a :class:`fiftyone.core.view.DatasetView`
        """
        if self.unique_ids is None:
            raise ValueError(
                "You must first call `find_unique()` or `find_duplicates()` "
                "to generate results"
            )

        samples = self.view
        patches_field = self.config.patches_field
        unique_ids = self.unique_ids

        if patches_field is not None and not isinstance(
            samples, fop.PatchesView
        ):
            samples = samples.to_patches(patches_field)

        return samples.select(unique_ids)

    def visualize_duplicates(self, visualization, backend="plotly", **kwargs):
        """Generates an interactive scatterplot of the results generated by the
        last call to :meth:`find_duplicates`.

        The ``visualization`` argument can be any visualization computed on the
        same dataset (or subset of it) as long as it contains every
        sample/object in the view whose results you are visualizing.

        The points are colored based on the following partition:

            -   "duplicate": duplicate example
            -   "nearest": nearest neighbor of a duplicate example
            -   "unique": the remaining unique examples

        Edges are also drawn between each duplicate and its nearest
        non-duplicate neighbor.

        You can attach plots generated by this method to an App session via its
        :attr:`fiftyone.core.session.Session.plots` attribute, which will
        automatically sync the session's view with the currently selected
        points in the plot.

        Args:
            visualization: a
                :class:`fiftyone.brain.visualization.VisualizationResults`
                instance to use to visualize the results
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method:

                -   "plotly" backend: :meth:`fiftyone.core.plots.plotly.scatterplot`
                -   "matplotlib" backend: :meth:`fiftyone.core.plots.matplotlib.scatterplot`

        Returns:
            a :class:`fiftyone.core.plots.base.InteractivePlot`
        """
        if self.neighbors_map is None:
            raise ValueError(
                "You must first call `find_duplicates()` to generate results"
            )

        samples = self.view
        duplicate_ids = self.duplicate_ids
        neighbors_map = self.neighbors_map
        patches_field = self.config.patches_field

        dup_ids = set(duplicate_ids)
        nearest_ids = set(neighbors_map.keys())

        with visualization.use_view(samples, allow_missing=True):
            if patches_field is not None:
                ids = visualization.current_label_ids
            else:
                ids = visualization.current_sample_ids

            labels = []
            for _id in ids:
                if _id in dup_ids:
                    label = "duplicate"
                elif _id in nearest_ids:
                    label = "nearest"
                else:
                    label = "unique"

                labels.append(label)

            if backend == "plotly":
                kwargs["edges"] = _build_edges(ids, neighbors_map)
                kwargs["edges_title"] = "neighbors"
                kwargs["labels_title"] = "type"

            return visualization.visualize(
                labels=labels,
                classes=["unique", "nearest", "duplicate"],
                backend=backend,
                **kwargs,
            )

    def visualize_unique(self, visualization, backend="plotly", **kwargs):
        """Generates an interactive scatterplot of the results generated by the
        last call to :meth:`find_unique`.

        The ``visualization`` argument can be any visualization computed on the
        same dataset (or subset of it) as long as it contains every
        sample/object in the view whose results you are visualizing.

        The points are colored based on the following partition:

            -   "unique": the unique examples
            -   "other": the other examples

        You can attach plots generated by this method to an App session via its
        :attr:`fiftyone.core.session.Session.plots` attribute, which will
        automatically sync the session's view with the currently selected
        points in the plot.

        Args:
            visualization: a
                :class:`fiftyone.brain.visualization.VisualizationResults`
                instance to use to visualize the results
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method:

                -   "plotly" backend: :meth:`fiftyone.core.plots.plotly.scatterplot`
                -   "matplotlib" backend: :meth:`fiftyone.core.plots.matplotlib.scatterplot`

        Returns:
            a :class:`fiftyone.core.plots.base.InteractivePlot`
        """
        if self.unique_ids is None:
            raise ValueError(
                "You must first call `find_unique()` to generate results"
            )

        samples = self.view
        unique_ids = self.unique_ids
        patches_field = self.config.patches_field

        unique_ids = set(unique_ids)

        with visualization.use_view(samples, allow_missing=True):
            if patches_field is not None:
                ids = visualization.current_label_ids
            else:
                ids = visualization.current_sample_ids

            labels = []
            for _id in ids:
                if _id in unique_ids:
                    label = "unique"
                else:
                    label = "other"

                labels.append(label)

            return visualization.visualize(
                labels=labels,
                classes=["other", "unique"],
                backend=backend,
                **kwargs,
            )


def _unique_no_sort(values):
    seen = set()
    return [v for v in values if v not in seen and not seen.add(v)]


def _build_edges(ids, neighbors_map):
    inds_map = {_id: idx for idx, _id in enumerate(ids)}

    edges = []
    for nearest_id, duplicates in neighbors_map.items():
        nearest_ind = inds_map[nearest_id]
        for dup_id, _ in duplicates:
            dup_ind = inds_map[dup_id]
            edges.append((dup_ind, nearest_ind))

    return np.array(edges)


def _plot_distances_plotly(dists, metric, thresh, bins, log, **kwargs):
    import plotly.graph_objects as go
    import fiftyone.core.plots.plotly as fopl

    counts, edges = np.histogram(dists, bins=bins)
    left_edges = edges[:-1]
    widths = edges[1:] - edges[:-1]
    customdata = np.stack((edges[:-1], edges[1:]), axis=1)

    hover_lines = [
        "<b>count: %{y}</b>",
        "distance: [%{customdata[0]:.2f}, %{customdata[1]:.2f}]",
    ]
    hovertemplate = "<br>".join(hover_lines) + "<extra></extra>"

    bar = go.Bar(
        x=left_edges,
        y=counts,
        width=widths,
        customdata=customdata,
        offset=0,
        marker_color="#FF6D04",
        hovertemplate=hovertemplate,
        showlegend=False,
    )

    traces = [bar]

    if thresh is not None:
        line = go.Scatter(
            x=[thresh, thresh],
            y=[0, max(counts)],
            mode="lines",
            line=dict(color="#17191C", width=3),
            hovertemplate="<b>thresh: %{x}</b><extra></extra>",
            showlegend=False,
        )
        traces.append(line)

    figure = go.Figure(traces)

    figure.update_layout(
        xaxis_title="nearest neighbor distance (%s)" % metric,
        yaxis_title="count",
        hovermode="x",
        yaxis_rangemode="tozero",
    )

    if log:
        figure.update_layout(yaxis_type="log")

    figure.update_layout(**fopl._DEFAULT_LAYOUT)
    figure.update_layout(**kwargs)

    if foc.is_jupyter_context():
        figure = fopl.PlotlyNotebookPlot(figure)

    return figure


def _plot_distances_mpl(
    dists, metric, thresh, bins, log, ax=None, figsize=None, **kwargs
):
    import matplotlib.pyplot as plt

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    counts, edges = np.histogram(dists, bins=bins)
    left_edges = edges[:-1]
    widths = edges[1:] - edges[:-1]

    ax.bar(
        left_edges,
        counts,
        width=widths,
        align="edge",
        color="#FF6D04",
        **kwargs,
    )

    if thresh is not None:
        ax.vlines(thresh, 0, max(counts), color="#17191C", linewidth=3)

    if log:
        ax.set_yscale("log")

    ax.set_xlabel("nearest neighbor distance (%s)" % metric)
    ax.set_ylabel("count")

    if figsize is not None:
        fig.set_size_inches(*figsize)

    plt.tight_layout()

    return fig



"""
Brain config.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os

from fiftyone.core.config import EnvConfig


class BrainConfig(EnvConfig):
    """FiftyOne brain configuration settings."""

    _BUILTIN_SIMILARITY_BACKENDS = {
        "sklearn": {
            "config_cls": "fiftyone.brain.internal.core.sklearn.SklearnSimilarityConfig",
        },
        "pinecone": {
            "config_cls": "fiftyone.brain.internal.core.pinecone.PineconeSimilarityConfig",
        },
        "qdrant": {
            "config_cls": "fiftyone.brain.internal.core.qdrant.QdrantSimilarityConfig",
        },
        "milvus": {
            "config_cls": "fiftyone.brain.internal.core.milvus.MilvusSimilarityConfig",
        },
        "lancedb": {
            "config_cls": "fiftyone.brain.internal.core.lancedb.LanceDBSimilarityConfig",
        },
        "redis": {
            "config_cls": "fiftyone.brain.internal.core.redis.RedisSimilarityConfig",
        },
        "mongodb": {
            "config_cls": "fiftyone.brain.internal.core.mongodb.MongoDBSimilarityConfig",
        },
        "elasticsearch": {
            "config_cls": "fiftyone.brain.internal.core.elasticsearch.ElasticsearchSimilarityConfig",
        },
        "pgvector": {
            "config_cls": "fiftyone.brain.internal.core.pgvector.PgVectorSimilarityConfig",
        },

        "mosaic": {
            "config_cls": "fiftyone.brain.internal.core.mosaic.MosaicSimilarityConfig",
        },
    }

    _BUILTIN_VISUALIZATION_METHODS = {
        "umap": {
            "config_cls": "fiftyone.brain.visualization.UMAPVisualizationConfig",
        },
        "tsne": {
            "config_cls": "fiftyone.brain.visualization.TSNEVisualizationConfig",
        },
        "pca": {
            "config_cls": "fiftyone.brain.visualization.PCAVisualizationConfig",
        },
        "manual": {
            "config_cls": "fiftyone.brain.visualization.ManualVisualizationConfig",
        },
    }

    def __init__(self, d=None):
        if d is None:
            d = {}

        self.default_similarity_backend = self.parse_string(
            d,
            "default_similarity_backend",
            env_var="FIFTYONE_BRAIN_DEFAULT_SIMILARITY_BACKEND",
            default="sklearn",
        )

        self.similarity_backends = self._parse_similarity_backends(d)
        if self.default_similarity_backend not in self.similarity_backends:
            self.default_similarity_backend = next(
                iter(sorted(self.similarity_backends.keys())), None
            )

        self.default_visualization_method = self.parse_string(
            d,
            "default_visualization_method",
            env_var="FIFTYONE_BRAIN_DEFAULT_VISUALIZATION_METHOD",
            default="umap",
        )

        self.visualization_methods = self._parse_visualization_methods(d)
        if self.default_visualization_method not in self.visualization_methods:
            self.default_visualization_method = next(
                iter(sorted(self.visualization_methods.keys())), None
            )

    def _parse_similarity_backends(self, d):
        d = d.get("similarity_backends", {})
        env_vars = dict(os.environ)

        #
        # `FIFTYONE_BRAIN_SIMILARITY_BACKENDS` can be used to declare which
        # backends are exposed. This may exclude builtin backends and/or
        # declare new backends
        #

        if "FIFTYONE_BRAIN_SIMILARITY_BACKENDS" in env_vars:
            backends = env_vars["FIFTYONE_BRAIN_SIMILARITY_BACKENDS"].split(
                ","
            )

            # Special syntax to append rather than override default backends
            if "*" in backends:
                backends = set(b for b in backends if b != "*")
                backends |= set(self._BUILTIN_SIMILARITY_BACKENDS.keys())

            d = {backend: d.get(backend, {}) for backend in backends}
        else:
            backends = self._BUILTIN_SIMILARITY_BACKENDS.keys()
            for backend in backends:
                if backend not in d:
                    d[backend] = {}

        #
        # Extract parameters from any environment variables of the form
        # `FIFTYONE_BRAIN_SIMILARITY_<BACKEND>_<PARAMETER>`
        #

        for backend, d_backend in d.items():
            prefix = "FIFTYONE_BRAIN_SIMILARITY_%s_" % backend.upper()
            for env_name, env_value in env_vars.items():
                if env_name.startswith(prefix):
                    name = env_name[len(prefix) :].lower()
                    value = _parse_env_value(env_value)
                    d_backend[name] = value

        #
        # Set default parameters for builtin similarity backends
        #

        for backend, defaults in self._BUILTIN_SIMILARITY_BACKENDS.items():
            if backend not in d:
                continue

            d_backend = d[backend]
            for name, value in defaults.items():
                if name not in d_backend:
                    d_backend[name] = value

        return d

    def _parse_visualization_methods(self, d):
        d = d.get("visualization_methods", {})
        env_vars = dict(os.environ)

        #
        # `FIFTYONE_BRAIN_VISUALIZATION_METHODS` can be used to declare which
        # methods are exposed. This may exclude builtin methods and/or declare
        # new methods
        #

        if "FIFTYONE_BRAIN_VISUALIZATION_METHODS" in env_vars:
            methods = env_vars["FIFTYONE_BRAIN_VISUALIZATION_METHODS"].split(
                ","
            )

            # Special syntax to append rather than override default methods
            if "*" in methods:
                methods = set(m for m in methods if m != "*")
                methods |= set(self._BUILTIN_VISUALIZATION_METHODS.keys())

            d = {method: d.get(method, {}) for method in methods}
        else:
            methods = self._BUILTIN_VISUALIZATION_METHODS.keys()
            for method in methods:
                if method not in d:
                    d[method] = {}

        #
        # Extract parameters from any environment variables of the form
        # `FIFTYONE_BRAIN_VISUALIZATION_<METHOD>_<PARAMETER>`
        #

        for method, d_method in d.items():
            prefix = "FIFTYONE_BRAIN_VISUALIZATION_%s_" % method.upper()
            for env_name, env_value in env_vars.items():
                if env_name.startswith(prefix):
                    name = env_name[len(prefix) :].lower()
                    value = _parse_env_value(env_value)
                    d_method[name] = value

        #
        # Set default parameters for builtin visualization methods
        #

        for method, defaults in self._BUILTIN_VISUALIZATION_METHODS.items():
            if method not in d:
                continue

            d_method = d[method]
            for name, value in defaults.items():
                if name not in d_method:
                    d_method[name] = value

        return d


def locate_brain_config():
    """Returns the path to the :class:`BrainConfig` on disk.

    The default location is ``~/.fiftyone/brain_config.json``, but you can
    override this path by setting the ``FIFTYONE_BRAIN_CONFIG_PATH``
    environment variable.

    Note that a config file may not actually exist on disk.

    Returns:
        the path to the :class:`BrainConfig` on disk
    """
    if "FIFTYONE_BRAIN_CONFIG_PATH" not in os.environ:
        return os.path.join(
            os.path.expanduser("~"), ".fiftyone", "brain_config.json"
        )

    return os.environ["FIFTYONE_BRAIN_CONFIG_PATH"]


def load_brain_config():
    """Loads the FiftyOne brain config.

    Returns:
        a :class:`BrainConfig` instance
    """
    brain_config_path = locate_brain_config()
    if os.path.isfile(brain_config_path):
        return BrainConfig.from_json(brain_config_path)

    return BrainConfig()


def _parse_env_value(value):
    try:
        return int(value)
    except:
        pass

    try:
        return float(value)
    except:
        pass

    if value in ("True", "true"):
        return True

    if value in ("False", "false"):
        return False

    if value in ("None", ""):
        return None

    if "," in value:
        return [_parse_env_value(v) for v in value.split(",")]

    return value



"""
Visualization tests.

All of these tests are designed to be run manually via::

    pytest tests/intensive/test_visualization.py -s -k test_<name>

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import unittest

import cv2
import numpy as np
import random as rand

import fiftyone as fo
import fiftyone.brain as fob
import fiftyone.zoo as foz
from fiftyone import ViewField as F


def test_mnist():
    dataset = foz.load_zoo_dataset("mnist", split="test")

    # pylint: disable=no-member
    embeddings = np.array(
        [
            cv2.imread(f, cv2.IMREAD_UNCHANGED).ravel()
            for f in dataset.values("filepath")
        ]
    )

    results = fob.compute_visualization(
        dataset,
        embeddings=embeddings,
        num_dims=2,
        verbose=True,
        seed=51,
    )

    plot = results.visualize(labels="ground_truth.label")
    plot.show()

    input("Press enter to continue...")


def test_images():
    dataset = _load_images_dataset()

    results = dataset.load_brain_results("img_viz")

    assert results.total_index_size == len(dataset)
    assert set(dataset.values("id")) == set(results.sample_ids)

    plot = results.visualize(labels="uniqueness")
    plot.show()

    input("Press enter to continue...")


def test_images_subset():
    dataset = _load_images_dataset()

    results = dataset.load_brain_results("img_viz")

    view = dataset.take(10)
    results.use_view(view)

    assert results.index_size == len(view)
    assert set(view.values("id")) == set(results.current_sample_ids)

    plot = results.visualize(labels="uniqueness")
    plot.show()

    input("Press enter to continue...")


def test_images_missing():
    dataset = _load_images_dataset().limit(4).clone()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    sample_ids = dataset[:4].values("id")

    results = fob.compute_visualization(dataset, batch_size=1)

    assert results.total_index_size == 4
    assert set(sample_ids) == set(results.sample_ids)

    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    results = fob.compute_visualization(
        dataset,
        model=model,
        embeddings="embeddings_missing",
        batch_size=1,
    )

    assert len(dataset.exists("embeddings_missing")) == 4
    assert results.total_index_size == 4
    assert set(sample_ids) == set(results.sample_ids)


def test_patches():
    dataset = _load_patches_dataset()

    results = dataset.load_brain_results("gt_viz")

    label_ids = dataset.values("ground_truth.detections.id", unwind=True)

    assert results.total_index_size == len(label_ids)
    assert set(label_ids) == set(results.label_ids)

    plot = results.visualize(labels="ground_truth.detections.label")
    plot.show()

    input("Press enter to continue...")


def test_patches_subset():
    dataset = _load_patches_dataset()

    results = dataset.load_brain_results("gt_viz")

    plot = results.visualize(
        labels="ground_truth.detections.label",
        classes=["person"],
    )
    plot.show()

    input("Press enter to continue...")

    view = dataset.filter_labels("ground_truth", F("label") == "person")
    results.use_view(view)

    label_ids = view.values("ground_truth.detections.id", unwind=True)

    assert results.index_size == len(label_ids)
    assert set(label_ids) == set(results.current_label_ids)

    plot = results.visualize(labels="ground_truth.detections.label")
    plot.show()

    input("Press enter to continue...")


def test_patches_missing():
    dataset = _load_patches_dataset().limit(4).clone()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    for sample in dataset[4:]:
        sample["ground_truth"] = fo.Detections(
            detections=[fo.Detection(bounding_box=[0.1, 0.1, 0.8, 0.8])]
        )
        sample.save()

    results = fob.compute_visualization(
        dataset, patches_field="ground_truth", batch_size=1
    )

    num_patches = dataset[:4].count("ground_truth.detections")
    label_ids = dataset[:4].values("ground_truth.detections.id", unwind=True)

    assert results.total_index_size == num_patches
    assert set(label_ids) == set(results.label_ids)

    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    results = fob.compute_visualization(
        dataset,
        model=model,
        patches_field="ground_truth",
        embeddings="embeddings_missing",
        batch_size=1,
    )

    view = dataset.filter_labels(
        "ground_truth", F("embeddings_missing") != None
    )

    assert view.count("ground_truth.detections") == num_patches
    assert results.total_index_size == num_patches
    assert set(label_ids) == set(results.label_ids)


def test_points():
    dataset = foz.load_zoo_dataset("quickstart")

    n = len(dataset)
    p = dataset.count("ground_truth.detections")
    d = 512

    points1 = np.random.rand(n, d)
    results1 = fob.compute_visualization(
        dataset,
        points=points1,
        brain_key="test1",
    )
    assert results1.points.shape == (n, d)

    points2 = {_id: np.random.rand(d) for _id in dataset.values("id")}
    results2 = fob.compute_visualization(
        dataset,
        points=points2,
        brain_key="test2",
    )
    assert results2.points.shape == (n, d)

    points3 = np.random.rand(p, d)
    results3 = fob.compute_visualization(
        dataset,
        patches_field="ground_truth",
        points=points3,
        brain_key="test3",
    )
    assert results3.points.shape == (p, d)

    points4 = {
        _id: np.random.rand(d)
        for _id in dataset.values("ground_truth.detections.id", unwind=True)
    }
    results4 = fob.compute_visualization(
        dataset,
        patches_field="ground_truth",
        points=points4,
        brain_key="test4",
    )
    assert results4.points.shape == (p, d)

    dataset.delete()


def test_similarity_index():
    dataset = foz.load_zoo_dataset(
        "quickstart", dataset_name=fo.get_default_dataset_name()
    )

    # Full similarity index

    similarity_index = fob.compute_similarity(
        dataset, brain_key="sklearn_index", backend="sklearn"
    )

    results = fob.compute_visualization(
        dataset,
        brain_key="img_viz",
        similarity_index=similarity_index,
    )

    assert len(results.points) == len(dataset)

    # Partial similarity index

    view = dataset.take(100, seed=51)
    similarity_index2 = fob.compute_similarity(
        view, brain_key="sklearn_index2", backend="sklearn"
    )

    results2 = fob.compute_visualization(
        dataset,
        brain_key="img_viz2",
        similarity_index="sklearn_index2",
    )

    assert len(results2.points) == len(view)


def test_points_field():
    dataset = _load_images_dataset()

    num_points = len(dataset)
    points = np.random.randn(num_points, 2)

    brain_key = "test_points"
    points_field = brain_key

    fob.compute_visualization(
        dataset,
        brain_key=brain_key,
        points=points,
        create_index=True,
    )

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field == points_field
    assert dataset.has_sample_field(points_field)
    assert points_field in dataset.list_indexes()

    sample_points = dataset.first()[points_field]

    assert isinstance(sample_points, list)
    assert len(sample_points) == 2
    assert isinstance(sample_points[0], float)

    points = results.points

    assert len(points) == num_points
    assert len(points[0]) == 2

    all_points = dataset.values(points_field)

    assert np.allclose(points, all_points)

    dataset.delete_brain_run(brain_key)

    assert not dataset.has_sample_field(points_field)
    assert points_field not in dataset.list_indexes()


def test_points_field_patches():
    dataset = _load_patches_dataset()

    num_points = dataset.count("ground_truth.detections")
    points = np.random.randn(num_points, 2)

    brain_key = "test_points"
    points_field = brain_key
    points_path = f"ground_truth.detections.{points_field}"

    fob.compute_visualization(
        dataset,
        brain_key=brain_key,
        points=points,
        patches_field="ground_truth",
        create_index=True,
    )

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field == points_field
    assert dataset.has_sample_field(points_path)
    # Patch visualizations can't currently make use of database indexes
    assert points_path not in dataset.list_indexes()

    label_points = dataset.first().ground_truth.detections[0][points_field]

    assert isinstance(label_points, list)
    assert len(label_points) == 2
    assert isinstance(label_points[0], float)

    points = results.points

    assert len(points) == num_points
    assert len(points[0]) == 2

    all_points = dataset.values(f"ground_truth.detections[].{points_field}")

    assert np.allclose(points, all_points)

    dataset.delete_brain_run(brain_key)

    assert not dataset.has_sample_field(points_path)


def test_index_points():
    dataset = _load_images_dataset()

    num_points = len(dataset)
    points = np.random.randn(num_points, 2)

    brain_key = "test_points"
    points_field = brain_key

    fob.compute_visualization(dataset, brain_key=brain_key, points=points)

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field is None
    assert not dataset.has_sample_field(points_field)
    assert points_field not in dataset.list_indexes()

    results.index_points()

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field == points_field
    assert dataset.has_sample_field(points_field)
    assert points_field in dataset.list_indexes()

    points = results.points
    all_points = dataset.values(points_field)

    assert np.allclose(points, all_points)

    results.remove_index()

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field is None
    assert not dataset.has_sample_field(points_field)
    assert points_field not in dataset.list_indexes()


def test_index_points_patches():
    dataset = _load_patches_dataset()

    num_points = dataset.count("ground_truth.detections")
    points = np.random.randn(num_points, 2)

    brain_key = "test_points"
    points_field = brain_key
    points_path = f"ground_truth.detections.{points_field}"

    fob.compute_visualization(
        dataset,
        brain_key=brain_key,
        points=points,
        patches_field="ground_truth",
    )

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field is None
    assert not dataset.has_sample_field(points_path)

    results.index_points()

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field == points_field
    assert dataset.has_sample_field(points_path)

    points = results.points
    all_points = dataset.values(f"ground_truth.detections[].{points_field}")

    assert np.allclose(points, all_points)

    results.remove_index()

    dataset.clear_cache()
    results = dataset.load_brain_results(brain_key)

    assert results.config.points_field is None
    assert not dataset.has_sample_field(points_path)


def _load_images_dataset():
    name = "test-visualization-images"

    if fo.dataset_exists(name):
        return fo.load_dataset(name)

    return _make_images_dataset(name)


def _load_patches_dataset():
    name = "test-visualization-patches"

    if fo.dataset_exists(name):
        return fo.load_dataset(name)

    return _make_patches_dataset(name)


def _make_images_dataset(name):
    dataset = foz.load_zoo_dataset(
        "quickstart", max_samples=20, dataset_name=name
    )
    model = foz.load_zoo_model("inception-v3-imagenet-torch")

    # Embed images
    dataset.compute_embeddings(
        model, embeddings_field="embeddings", batch_size=8
    )

    # Image visualization
    fob.compute_visualization(
        dataset,
        embeddings="embeddings",
        num_dims=2,
        verbose=True,
        seed=51,
        brain_key="img_viz",
    )

    return dataset


def _make_patches_dataset(name):
    dataset = foz.load_zoo_dataset(
        "quickstart", max_samples=20, dataset_name=name
    )
    model = foz.load_zoo_model("inception-v3-imagenet-torch")

    # Embed ground truth patches
    dataset.compute_patch_embeddings(
        model,
        "ground_truth",
        embeddings_field="embeddings",
        batch_size=8,
        force_square=True,
    )

    # Patch visualization
    fob.compute_visualization(
        dataset,
        patches_field="ground_truth",
        embeddings="embeddings",
        num_dims=2,
        verbose=True,
        seed=51,
        brain_key="gt_viz",
    )

    return dataset


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)



"""
Milvus similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np
from uuid import uuid4

import eta.core.utils as etau

import fiftyone.core.utils as fou
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)
import fiftyone.brain.internal.core.utils as fbu

pymilvus = fou.lazy_import("pymilvus")


logger = logging.getLogger(__name__)

_SUPPORTED_METRICS = {
    "cosine": "COSINE",
    "dotproduct": "IP",
    "euclidean": "L2",
}


class MilvusSimilarityConfig(SimilarityConfig):
    """Configuration for the Milvus similarity backend.

    Args:
        collection_name (None): the name of a Milvus collection to use or
            create. If none is provided, a new collection will be created
        metric ("dotproduct"): the embedding distance metric to use when
            creating a new index. Supported values are
            ``("cosine", "dotproduct", "euclidean")``
        consistency_level ("Session"): the consistency level to use. Supported
            values are ``("Session", "Strong", "Bounded", "Eventually")``
        uri (None): a full Milvus server address to use, like
            ``"http://localhost:19530"``,
            ``"tcp:localhost:19530"``, or
            ``"https://ok.s3.south.com:19530"``
        user (None): a username to use
        password (None): a password to use
        secure (None): whether to enable TLS (True)
        token (None): a header token for RPC calls
        db_name (None): a database name for the connection
        client_key_path (None): a client.key path for TLS two-way
        client_pem_path (None): a client.pem path for TLS two-way
        ca_pem_path (None): a ca.pem path for TLS two-way
        server_pem_path (None): a server.pem path for TLS one-way
        server_name (None): the server name, for TLS
        **kwargs: keyword arguments for
            :class:`fiftyone.brain.similarity.SimilarityConfig`
    """

    def __init__(
        self,
        collection_name=None,
        metric="dotproduct",
        consistency_level="Session",
        uri=None,
        user=None,
        password=None,
        secure=None,
        token=None,
        db_name=None,
        client_key_path=None,
        client_pem_path=None,
        ca_pem_path=None,
        server_pem_path=None,
        server_name=None,
        **kwargs,
    ):
        if metric not in _SUPPORTED_METRICS:
            raise ValueError(
                "Unsupported metric '%s'. Supported values are %s"
                % (metric, tuple(_SUPPORTED_METRICS.keys()))
            )

        super().__init__(**kwargs)

        self.collection_name = collection_name
        self.metric = metric
        self.consistency_level = consistency_level

        # store privately so these aren't serialized
        self._uri = uri
        self._user = user
        self._password = password
        self._secure = secure
        self._token = token
        self._db_name = db_name
        self._client_key_path = client_key_path
        self._client_pem_path = client_pem_path
        self._ca_pem_path = ca_pem_path
        self._server_pem_path = server_pem_path
        self._server_name = server_name

    @property
    def method(self):
        return "milvus"

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, value):
        self._uri = value

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, value):
        self._user = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value

    @property
    def secure(self):
        return self._secure

    @secure.setter
    def secure(self, value):
        self._secure = value

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, value):
        self._token = value

    @property
    def db_name(self):
        return self._db_name

    @db_name.setter
    def db_name(self, value):
        self._db_name = value

    @property
    def client_key_path(self):
        return self._client_key_path

    @client_key_path.setter
    def client_key_path(self, value):
        self._client_key_path = value

    @property
    def client_pem_path(self):
        return self._client_pem_path

    @client_pem_path.setter
    def client_pem_path(self, value):
        self._client_pem_path = value

    @property
    def ca_pem_path(self):
        return self._ca_pem_path

    @ca_pem_path.setter
    def ca_pem_path(self, value):
        self._ca_pem_path = value

    @property
    def server_pem_path(self):
        return self._server_pem_path

    @server_pem_path.setter
    def server_pem_path(self, value):
        self._server_pem_path = value

    @property
    def server_name(self):
        return self._server_name

    @server_name.setter
    def server_name(self, value):
        self._server_name = value

    @property
    def max_k(self):
        return 16384

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    @property
    def index_params(self):
        return {
            "metric_type": _SUPPORTED_METRICS[self.metric],
            "index_type": "HNSW",
            "params": {"M": 8, "efConstruction": 64},
        }

    @property
    def search_params(self):
        return {
            "HNSW": {
                "metric_type": _SUPPORTED_METRICS[self.metric],
                "params": {"ef": 10},
            },
        }

    def load_credentials(
        self,
        uri=None,
        user=None,
        password=None,
        secure=None,
        token=None,
        db_name=None,
        client_key_path=None,
        client_pem_path=None,
        ca_pem_path=None,
        server_pem_path=None,
        server_name=None,
    ):
        self._load_parameters(
            uri=uri,
            user=user,
            password=password,
            secure=secure,
            token=token,
            db_name=db_name,
            client_key_path=client_key_path,
            client_pem_path=client_pem_path,
            ca_pem_path=ca_pem_path,
            server_pem_path=server_pem_path,
            server_name=server_name,
        )


class MilvusSimilarity(Similarity):
    """Milvus similarity factory.

    Args:
        config: a :class:`MilvusSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("pymilvus")

    def ensure_usage_requirements(self):
        fou.ensure_package("pymilvus")

    def initialize(self, samples, brain_key):
        return MilvusSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class MilvusSimilarityIndex(SimilarityIndex):
    """Class for interacting with Milvus similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`MilvusSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`MilvusSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._alias = None
        self._collection = None
        self._initialize()

    def _initialize(self):
        kwargs = {}

        for key in (
            "uri",
            "user",
            "password",
            "secure",
            "token",
            "db_name",
            "client_key_path",
            "client_pem_path",
            "ca_pem_path",
            "server_pem_path",
            "server_name",
        ):
            value = getattr(self.config, key, None)
            if value is not None:
                kwargs[key] = value

        alias = uuid4().hex if kwargs else "default"

        try:
            pymilvus.connections.connect(alias=alias, **kwargs)
        except pymilvus.MilvusException as e:
            raise ValueError(
                "Failed to connect to Milvus backend at URI '%s'. Refer to "
                "https://docs.voxel51.com/integrations/milvus.html for more "
                "information" % self.config.uri
            ) from e

        collection_names = pymilvus.utility.list_collections(using=alias)

        if self.config.collection_name is None:
            # Milvus only supports numbers, letters and underscores
            root = "fiftyone-" + fou.to_slug(self.samples._root_dataset.name)
            root = root.replace("-", "_")
            collection_name = fbu.get_unique_name(root, collection_names)
            collection_name = collection_name.replace("-", "_")

            self.config.collection_name = collection_name
            self.save_config()

        if self.config.collection_name in collection_names:
            collection = pymilvus.Collection(
                self.config.collection_name, using=alias
            )
            collection.load()
        else:
            collection = None

        self._alias = alias
        self._collection = collection

    def _create_collection(self, dimension):
        schema = pymilvus.CollectionSchema(
            [
                pymilvus.FieldSchema(
                    "pk",
                    pymilvus.DataType.VARCHAR,
                    is_primary=True,
                    auto_id=False,
                    max_length=64000,
                ),
                pymilvus.FieldSchema(
                    "vector", pymilvus.DataType.FLOAT_VECTOR, dim=dimension
                ),
                pymilvus.FieldSchema(
                    "sample_id", pymilvus.DataType.VARCHAR, max_length=64000
                ),
            ]
        )

        collection = pymilvus.Collection(
            self.config.collection_name,
            schema,
            consistency_level=self.config.consistency_level,
            using=self._alias,
        )
        collection.create_index(
            "vector", index_params=self.config.index_params
        )
        collection.load()

        self._collection = collection

    @property
    def collection(self):
        """The ``pymilvus.Collection`` instance for this index."""
        return self._collection

    @property
    def total_index_size(self):
        if self._collection is None:
            return 0

        return self._collection.num_entities

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
        batch_size=100,
    ):
        if self._collection is None:
            self._create_collection(embeddings.shape[1])

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            existing_ids = self._get_existing_ids(ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            del_inds = [i for i, _id in enumerate(ids) if _id in existing_ids]
            embeddings = np.delete(embeddings, del_inds, axis=0)
            sample_ids = np.delete(sample_ids, del_inds)
            if label_ids is not None:
                label_ids = np.delete(label_ids, del_inds)
        elif existing_ids and overwrite:
            self._delete_ids(existing_ids)

        embeddings = [e.tolist() for e in embeddings]
        sample_ids = list(sample_ids)
        ids = list(ids)

        for _embeddings, _ids, _sample_ids in zip(
            fou.iter_batches(embeddings, batch_size),
            fou.iter_batches(ids, batch_size),
            fou.iter_batches(sample_ids, batch_size),
        ):
            insert_data = [
                list(_ids),
                list(_embeddings),
                list(_sample_ids),
            ]
            self._collection.insert(insert_data)

        self._collection.flush()

        if reload:
            self.reload()

    def _get_existing_ids(self, ids):
        ids = ['"' + str(entry) + '"' for entry in ids]
        expr = f"""pk in [{','.join(ids)}]"""
        return self._collection.query(expr)

    def _delete_ids(self, ids):
        ids = ['"' + str(entry) + '"' for entry in ids]
        expr = f"""pk in [{','.join(ids)}]"""
        self._collection.delete(expr)
        self._collection.flush()

    def _get_embeddings(self, ids):
        ids = ['"' + str(entry) + '"' for entry in ids]
        expr = f"""pk in [{','.join(ids)}]"""
        return self._collection.query(
            expr, output_fields=["pk", "sample_id", "vector"]
        )

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if not allow_missing or warn_missing:
            existing_ids = self._get_existing_ids(ids)
            missing_ids = set(ids) - set(existing_ids)
            num_missing = len(missing_ids)

            if num_missing > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that are not present in the "
                        "index" % (num_missing, next(iter(missing_ids)))
                    )

                if warn_missing:
                    logger.warning(
                        "Ignoring %d IDs that are not present in the index",
                        num_missing,
                    )

                ids = existing_ids

        self._delete_ids(ids=ids)

        if reload:
            self.reload()

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        if sample_ids is not None and self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_sample_ids(sample_ids)
        elif self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_label_ids(label_ids)
        else:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_sample_embeddings(sample_ids)

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(embeddings)
        sample_ids = np.array(sample_ids)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return embeddings, sample_ids, label_ids

    def cleanup(self):
        pymilvus.utility.drop_collection(
            self.config.collection_name, using=self._alias
        )
        self._collection = None

    def _get_sample_embeddings(self, sample_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []

        if sample_ids is None:
            raise ValueError(
                "Milvus does not support retrieving all vectors in an index"
            )

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            response = self._get_embeddings(list(batch_ids))

            for r in response:
                found_embeddings.append(r["vector"])
                found_sample_ids.append(r["sample_id"])

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, None, missing_ids

    def _get_patch_embeddings_from_label_ids(self, label_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        if label_ids is None:
            raise ValueError(
                "Milvus does not support retrieving all vectors in an index"
            )

        for batch_ids in fou.iter_batches(label_ids, batch_size):
            response = self._get_embeddings(list(batch_ids))

            for r in response:
                found_embeddings.append(r["vector"])
                found_sample_ids.append(r["sample_id"])
                found_label_ids.append(r["pk"])

        missing_ids = list(set(label_ids) - set(found_label_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _get_patch_embeddings_from_sample_ids(
        self, sample_ids, batch_size=100
    ):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        query_vector = [0.0] * self._get_dimension()
        top_k = min(batch_size, self.config.max_k)

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            ids = ['"' + str(entry) + '"' for entry in batch_ids]
            expr = f"""pk in [{','.join(ids)}]"""
            response = self._collection.search(
                data=[query_vector],
                anns_field="vector",
                param=self.config.search_params,
                expr=expr,
                limit=top_k,
            )
            ids = [x.id for x in response[0]]
            response = self._get_embeddings(ids)
            for r in response:
                found_embeddings.append(r["vector"])
                found_sample_ids.append(r["sample_id"])
                found_label_ids.append(r["pk"])

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        if query is None:
            raise ValueError("Milvus does not support full index neighbors")

        if reverse is True:
            raise ValueError(
                "Milvus does not support least similarity queries"
            )

        if k is None or k > self.config.max_k:
            raise ValueError("Milvus requires k<=%s" % self.config.max_k)

        if aggregation not in (None, "mean"):
            raise ValueError("Unsupported aggregation '%s'" % aggregation)

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        if self.has_view:
            if self.config.patches_field is not None:
                index_ids = self.current_label_ids
            else:
                index_ids = self.current_sample_ids

            expr = ['"' + str(entry) + '"' for entry in index_ids]
            expr = f"""pk in [{','.join(expr)}]"""
        else:
            expr = None

        ids = []
        dists = []
        for q in query:
            response = self._collection.search(
                data=[q.tolist()],
                anns_field="vector",
                limit=k,
                expr=expr,
                param=self.config.search_params,
            )
            ids.append([r.id for r in response[0]])
            if return_dists:
                dists.append([r.score for r in response[0]])

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        response = self._get_embeddings(query_ids)
        query = np.array([x["vector"] for x in response])

        if single_query:
            query = query[0, :]

        return query

    def _get_dimension(self):
        if self._collection is None:
            return None

        for field in self._collection.describe()["fields"]:
            if field["name"] == "vector":
                return field["params"]["dim"]

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)



"""
Qdrant similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

import fiftyone.core.utils as fou
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)
import fiftyone.brain.internal.core.utils as fbu

qdrant = fou.lazy_import("qdrant_client")
qmodels = fou.lazy_import("qdrant_client.http.models")


logger = logging.getLogger(__name__)

_SUPPORTED_METRICS = {
    "cosine": qmodels.Distance.COSINE,
    "dotproduct": qmodels.Distance.DOT,
    "euclidean": qmodels.Distance.EUCLID,
}


class QdrantSimilarityConfig(SimilarityConfig):
    """Configuration for the Qdrant similarity backend.

    Args:
        collection_name (None): the name of a Qdrant collection to use or
            create. If none is provided, a new collection will be created
        metric (None): the embedding distance metric to use when creating a
            new index. Supported values are
            ``("cosine", "dotproduct", "euclidean")``
        replication_factor (None): an optional replication factor to use when
            creating a new index
        shard_number (None): an optional number of shards to use when creating
            a new index
        write_consistency_factor (None): an optional write consistsency factor
            to use when creating a new index
        hnsw_config (None): an optional dict of HNSW config parameters to use
            when creating a new index
        optimizers_config (None): an optional dict of optimizer parameters to
            use when creating a new index
        wal_config (None): an optional dict of WAL config parameters to use
            when creating a new index
        url (None): a Qdrant server URL to use
        api_key (None): a Qdrant API key to use
        grpc_port (None): Port of Qdrant gRPC interface
        prefer_grpc (None): If `true`, use gRPC interface when possible
        **kwargs: keyword arguments for
            :class:`fiftyone.brain.similarity.SimilarityConfig`
    """

    def __init__(
        self,
        collection_name=None,
        metric=None,
        replication_factor=None,
        shard_number=None,
        write_consistency_factor=None,
        hnsw_config=None,
        optimizers_config=None,
        wal_config=None,
        url=None,
        api_key=None,
        grpc_port=None,
        prefer_grpc=None,
        **kwargs,
    ):
        if metric is not None and metric not in _SUPPORTED_METRICS:
            raise ValueError(
                "Unsupported metric '%s'. Supported values are %s"
                % (metric, tuple(_SUPPORTED_METRICS.keys()))
            )

        super().__init__(**kwargs)

        self.collection_name = collection_name
        self.metric = metric
        self.replication_factor = replication_factor
        self.shard_number = shard_number
        self.write_consistency_factor = write_consistency_factor
        self.hnsw_config = hnsw_config
        self.optimizers_config = optimizers_config
        self.wal_config = wal_config

        # store privately so these aren't serialized
        self._url = url
        self._api_key = api_key
        self._grpc_port = grpc_port
        self._prefer_grpc = prefer_grpc

    @property
    def method(self):
        return "qdrant"

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        self._api_key = value

    @property
    def grpc_port(self):
        return self._grpc_port

    @grpc_port.setter
    def grpc_port(self, value):
        self._grpc_port = value

    @property
    def prefer_grpc(self):
        return self._prefer_grpc

    @prefer_grpc.setter
    def prefer_grpc(self, value):
        self._prefer_grpc = value

    @property
    def max_k(self):
        return None

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    def load_credentials(
        self, url=None, api_key=None, grpc_port=None, prefer_grpc=None
    ):
        self._load_parameters(
            url=url,
            api_key=api_key,
            grpc_port=grpc_port,
            prefer_grpc=prefer_grpc,
        )


class QdrantSimilarity(Similarity):
    """Qdrant similarity factory.

    Args:
        config: a :class:`QdrantSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("qdrant-client")

    def ensure_usage_requirements(self):
        fou.ensure_package("qdrant-client")

    def initialize(self, samples, brain_key):
        return QdrantSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class QdrantSimilarityIndex(SimilarityIndex):
    """Class for interacting with Qdrant similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`QdrantSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`QdrantSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._client = None
        self._initialize()

    def _initialize(self):
        # QdrantClient does not appear to like passing None as defaults
        grpc_port = (
            self.config.grpc_port
            if self.config.grpc_port is not None
            else 6334
        )
        prefer_grpc = (
            self.config.prefer_grpc
            if self.config.prefer_grpc is not None
            else False
        )

        self._client = qdrant.QdrantClient(
            url=self.config.url,
            api_key=self.config.api_key,
            grpc_port=grpc_port,
            prefer_grpc=prefer_grpc,
        )

        try:
            collection_names = self._get_collection_names()
        except Exception as e:
            raise ValueError(
                "Failed to connect to Qdrant backend at URL '%s'. Refer to "
                "https://docs.voxel51.com/integrations/qdrant.html for more "
                "information" % self.config.url
            ) from e

        if self.config.collection_name is None:
            root = "fiftyone-" + fou.to_slug(self.samples._root_dataset.name)
            collection_name = fbu.get_unique_name(root, collection_names)

            self.config.collection_name = collection_name
            self.save_config()

    def _get_collection_names(self):
        return [c.name for c in self._client.get_collections().collections]

    def _create_collection(self, dimension):
        if self.config.metric:
            metric = self.config.metric
        else:
            metric = "cosine"

        vectors_config = qmodels.VectorParams(
            size=dimension,
            distance=_SUPPORTED_METRICS[metric],
        )

        if self.config.hnsw_config:
            hnsw_config = qmodels.HnswConfig(**self.config.hnsw_config)
        else:
            hnsw_config = None

        if self.config.optimizers_config:
            optimizers_config = qmodels.OptimizersConfig(
                **self.config.optimizers_config
            )
        else:
            optimizers_config = None

        if self.config.wal_config:
            wal_config = qmodels.WalConfig(**self.config.wal_config)
        else:
            wal_config = None

        self._client.recreate_collection(
            collection_name=self.config.collection_name,
            vectors_config=vectors_config,
            shard_number=self.config.shard_number,
            replication_factor=self.config.replication_factor,
            hnsw_config=hnsw_config,
            optimizers_config=optimizers_config,
            wal_config=wal_config,
        )

    def _get_index_ids(self, batch_size=1000):
        ids = []

        offset = 0
        while offset is not None:
            response = self._client.scroll(
                collection_name=self.config.collection_name,
                offset=offset,
                limit=batch_size,
                with_payload=True,
                with_vectors=False,
            )
            ids.extend([self._to_fiftyone_id(r.id) for r in response[0]])
            offset = response[-1]

        return ids

    @property
    def total_index_size(self):
        try:
            return self._client.count(self.config.collection_name).count
        except:
            return 0

    @property
    def client(self):
        """The ``qdrant.QdrantClient`` instance for this index."""
        return self._client

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
        batch_size=1000,
    ):
        if self.config.collection_name not in self._get_collection_names():
            self._create_collection(embeddings.shape[1])

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            index_ids = self._get_index_ids()

            existing_ids = set(ids) & set(index_ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            del_inds = [i for i, _id in enumerate(ids) if _id in existing_ids]
            embeddings = np.delete(embeddings, del_inds, axis=0)
            sample_ids = np.delete(sample_ids, del_inds)
            if label_ids is not None:
                label_ids = np.delete(label_ids, del_inds)

        embeddings = [e.tolist() for e in embeddings]
        sample_ids = list(sample_ids)
        if label_ids is not None:
            ids = list(label_ids)
        else:
            ids = list(sample_ids)

        for _embeddings, _ids, _sample_ids in zip(
            fou.iter_batches(embeddings, batch_size),
            fou.iter_batches(ids, batch_size),
            fou.iter_batches(sample_ids, batch_size),
        ):
            self._client.upsert(
                collection_name=self.config.collection_name,
                points=qmodels.Batch(
                    ids=self._to_qdrant_ids(_ids),
                    payloads=[{"sample_id": _id} for _id in _sample_ids],
                    vectors=_embeddings,
                ),
            )

        if reload:
            self.reload()

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        qids = self._to_qdrant_ids(ids)

        if warn_missing or not allow_missing:
            response = self._retrieve_points(qids, with_vectors=False)
            existing_ids = self._to_fiftyone_ids([r.id for r in response])
            missing_ids = set(ids) - set(existing_ids)
            num_missing_ids = len(missing_ids)

            if num_missing_ids > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that do not exist in the index"
                        % (num_missing_ids, next(iter(missing_ids)))
                    )
                if warn_missing and not allow_missing:
                    logger.warning(
                        "Skipping %d IDs that do not exist in the index",
                        num_missing_ids,
                    )

            qids = self._to_qdrant_ids(existing_ids)

        self._client.delete(
            collection_name=self.config.collection_name,
            points_selector=qmodels.PointIdsList(points=qids),
        )

        if reload:
            self.reload()

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        if sample_ids is not None and self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_sample_ids(sample_ids)
        elif self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_label_ids(label_ids)
        else:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_sample_embeddings(sample_ids)

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(embeddings)
        sample_ids = np.array(sample_ids)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return embeddings, sample_ids, label_ids

    def cleanup(self):
        self._client.delete_collection(self.config.collection_name)

    def _retrieve_points(self, qids, with_vectors=True, with_payload=True):
        # @todo add batching?
        return self._client.retrieve(
            collection_name=self.config.collection_name,
            ids=qids,
            with_vectors=with_vectors,
            with_payload=with_payload,
        )

    def _get_sample_embeddings(self, sample_ids):
        if sample_ids is None:
            sample_ids = self._get_index_ids()

        response = self._retrieve_points(
            self._to_qdrant_ids(sample_ids),
            with_vectors=True,
        )

        found_embeddings = [r.vector for r in response]
        found_sample_ids = self._to_fiftyone_ids([r.id for r in response])
        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, None, missing_ids

    def _get_patch_embeddings_from_label_ids(self, label_ids):
        if label_ids is None:
            label_ids = self._get_index_ids()

        response = self._retrieve_points(
            self._to_qdrant_ids(label_ids),
            with_vectors=True,
            with_payload=True,
        )

        found_embeddings = [r.vector for r in response]
        found_sample_ids = [r.payload["sample_id"] for r in response]
        found_label_ids = self._to_fiftyone_ids([r.id for r in response])
        missing_ids = list(set(label_ids) - set(found_label_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _get_patch_embeddings_from_sample_ids(self, sample_ids):
        _filter = qmodels.Filter(
            should=[
                qmodels.FieldCondition(
                    key="sample_id", match=qmodels.MatchValue(value=sid)
                )
                for sid in sample_ids
            ]
        )

        response = self._client.scroll(
            collection_name=self.config.collection_name,
            scroll_filter=_filter,
            with_vectors=True,
            with_payload=True,
        )[0]

        found_embeddings = [r.vector for r in response]
        found_sample_ids = [r.payload["sample_id"] for r in response]
        found_label_ids = [r.id for r in response]
        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        if query is None:
            raise ValueError("Qdrant does not support full index neighbors")

        if reverse is True:
            raise ValueError(
                "Qdrant does not support least similarity queries"
            )

        if aggregation not in (None, "mean"):
            raise ValueError("Unsupported aggregation '%s'" % aggregation)

        if k is None:
            k = self.index_size

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        if self.has_view:
            if self.config.patches_field is not None:
                index_ids = self.current_label_ids
            else:
                index_ids = self.current_sample_ids

            _filter = qmodels.Filter(
                must=[
                    qmodels.HasIdCondition(
                        has_id=self._to_qdrant_ids(index_ids)
                    )
                ]
            )
        else:
            _filter = None

        ids = []
        dists = []
        for q in query:
            results = self._client.search(
                collection_name=self.config.collection_name,
                query_vector=q,
                query_filter=_filter,
                with_payload=False,
                limit=k,
            )

            ids.append(self._to_fiftyone_ids([r.id for r in results]))
            if return_dists:
                dists.append([r.score for r in results])

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        qids = self._to_qdrant_ids(query_ids)
        response = self._retrieve_points(qids, with_vectors=True)
        query = np.array([r.vector for r in response])

        if single_query:
            query = query[0, :]

        return query

    def _to_qdrant_id(self, _id):
        return _id + "00000000"

    def _to_qdrant_ids(self, ids):
        return [self._to_qdrant_id(_id) for _id in ids]

    def _to_fiftyone_id(self, qid):
        return qid.replace("-", "")[:-8]

    def _to_fiftyone_ids(self, qids):
        return [self._to_fiftyone_id(qid) for qid in qids]

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)



"""
PyTorch utilities.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import fiftyone.utils.torch as fout

from fiftyone.brain.internal.models import HasBrainModel

import torch


class TorchImageModelConfig(fout.TorchImageModelConfig, HasBrainModel):
    """Configuration for running a :class:`TorchImageModel`.

    See :class:`fiftyone.utils.torch.TorchImageModelConfig` for additional
    parameters.

    Args:
        model_name (None): the name of the Brain model state dict to load
        model_path (None): the path to a state dict on disk to load
    """

    def __init__(self, d):
        d = self.init(d)
        super().__init__(d)


class TorchImageModel(fout.TorchImageModel):
    """Wrapper for evaluating a Torch model on images whose state dict is
    stored privately by the Brain.

    Args:
        config: an :class:`TorchImageModelConfig`
    """

    def _download_model(self, config):
        config.download_model_if_necessary()

    def _load_state_dict(self, model, config):
        state_dict = torch.load(config.model_path, map_location=self.device)
        model.load_state_dict(state_dict)



"""
MongoDB similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

from bson import ObjectId
import numpy as np
from pymongo.errors import OperationFailure

import eta.core.utils as etau

from fiftyone import ViewField as F
import fiftyone.core.fields as fof
import fiftyone.core.media as fom
import fiftyone.core.utils as fou
import fiftyone.brain.internal.core.utils as fbu
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)


logger = logging.getLogger(__name__)

_SUPPORTED_METRICS = {
    "cosine": "cosine",
    "dotproduct": "dotProduct",
    "euclidean": "euclidean",
}


class MongoDBSimilarityConfig(SimilarityConfig):
    """Configuration for a MongoDB similarity instance.

    Args:
        index_name (None): the name of the MongoDB vector index to use or
            create. If none is provided, a new index will be created
        metric ("cosine"): the embedding distance metric to use when creating a
            new index. Supported values are
            ``("cosine", "dotproduct", "euclidean")``
        **kwargs: keyword arguments for
            :class:`fiftyone.brain.similarity.SimilarityConfig`
    """

    def __init__(self, index_name=None, metric="cosine", **kwargs):
        if kwargs.get("embeddings_field") is None and index_name is None:
            raise ValueError(
                "You must provide either the name of a field to read/write "
                "embeddings for this index by passing the `embeddings` "
                "parameter, or you must provide the name of an existing "
                "vector search index via the `index_name` parameter"
            )

        # @todo support this. Will likely require copying embeddings to a new
        # collection as vector search indexes do not yet support array fields
        if kwargs.get("patches_field") is not None:
            raise ValueError(
                "The MongoDB backend does not yet support patch embeddings"
            )

        if metric not in _SUPPORTED_METRICS:
            raise ValueError(
                "Unsupported metric '%s'. Supported values are %s"
                % (metric, tuple(_SUPPORTED_METRICS.keys()))
            )

        super().__init__(**kwargs)

        self.index_name = index_name
        self.metric = metric

    @property
    def method(self):
        return "mongodb"

    @property
    def max_k(self):
        return 10000  # MongoDB limit

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)


class MongoDBSimilarity(Similarity):
    """MongoDB similarity factory.

    Args:
        config: a :class:`MongoDBSimilarityConfig`
    """

    def ensure_requirements(self):
        #
        # https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html#pymongo.collection.Collection.create_search_index
        #
        # Could also validate that user is connected to an Atlas cluster here
        # eg Atlas clusters generally have hostnames which end in "mongodb.net"
        # https://stackoverflow.com/q/73180110
        #
        fou.ensure_package("pymongo>=4.7")

    def ensure_usage_requirements(self):
        fou.ensure_package("pymongo>=4.7")

    def initialize(self, samples, brain_key):
        return MongoDBSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class MongoDBSimilarityIndex(SimilarityIndex):
    """Class for interacting with MongoDB similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`MongoDBSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`MongoDBSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        dataset = samples._dataset
        sample_ids, label_ids = self._parse_data(dataset, config)

        self._dataset = dataset
        self._sample_ids = sample_ids
        self._label_ids = label_ids

        super().__init__(samples, config, brain_key, backend=backend)

        self._index = None
        self._initialize()

    @property
    def is_external(self):
        return False

    @property
    def sample_ids(self):
        return self._sample_ids

    @property
    def label_ids(self):
        return self._label_ids

    @property
    def total_index_size(self):
        return len(self._sample_ids)

    def _initialize(self):
        coll = self._dataset._sample_collection

        try:
            indexes = {
                i["name"]: i
                for i in coll.aggregate([{"$listSearchIndexes": {}}])
            }
        except OperationFailure:
            # https://www.mongodb.com/docs/manual/release-notes/7.0/#atlas-search-index-management
            # https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-overview
            if self.config.index_name is None:
                raise ValueError(
                    "You must be running MongoDB Atlas 7.0 or later in order "
                    "to use vector search indexes"
                )

            # Must assume index exists because we can't use pymongo to check...
            self._index = True

            return

        if self.config.index_name is None:
            root = self.config.embeddings_field
            index_name = fbu.get_unique_name(root, list(indexes.keys()))

            self.config.index_name = index_name
            self.save_config()
        elif self.config.embeddings_field is None:
            info = indexes.get(self.config.index_name, None)
            if info is None:
                raise ValueError(
                    "Index '%s' does not exist" % self.config.index_name
                )

            self.config.embeddings_field = next(
                iter(info["latestDefinition"]["mappings"]["fields"].keys())
            )
            self.save_config()

        if self.config.index_name in indexes:
            # Index already exists
            self._index = True
        elif self.total_index_size > 0:
            # Embeddings already exist but the index hasn't been declared yet
            dimension = self._get_dimension()
            self._create_index(dimension)
        else:
            # Index will be created when add_to_index() is called
            pass

    def _get_dimension(self):
        if self.total_index_size == 0:
            return None

        if self.config.patches_field is not None:
            embeddings, _, _ = self.get_embeddings(
                label_ids=self._label_ids[:1]
            )
        else:
            embeddings, _, _ = self.get_embeddings(
                sample_ids=self._sample_ids[:1]
            )

        return embeddings.shape[1]

    def _create_index(self, dimension):
        # https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-stage
        # https://www.mongodb.com/docs/languages/python/pymongo-driver/current/indexes/atlas-search-index/
        from pymongo.operations import SearchIndexModel

        field = self._dataset.get_field(self.config.embeddings_field)
        if field is not None and not isinstance(field, fof.ListField):
            raise ValueError(
                "MongoDB vector search indexes require embeddings to be "
                "stored in list fields"
            )

        metric = _SUPPORTED_METRICS[self.config.metric]

        fields = [
            {
                "type": "vector",
                "numDimensions": dimension,
                "path": self.config.embeddings_field,
                "similarity": metric,
            }
        ]

        if self._dataset.media_type == fom.GROUP:
            fields.append(
                {
                    "type": "filter",
                    "path": self._dataset.group_field + ".name",
                }
            )

        model = SearchIndexModel(
            name=self.config.index_name,
            type="vectorSearch",  # requires pymongo>=4.7
            definition={"fields": fields},
        )

        coll = self._dataset._sample_collection
        coll.create_search_index(model=model)

        self._index = True

    @property
    def ready(self):
        """Returns True/False whether the vector search index is ready to be
        queried.
        """
        if self._index is None:
            return False

        try:
            coll = self._dataset._sample_collection
            indexes = {
                i["name"]: i
                for i in coll.aggregate([{"$listSearchIndexes": {}}])
            }
        except OperationFailure:
            # requires MongoDB Atlas 7.0 or later
            return None

        info = indexes.get(self.config.index_name, {})
        return info.get("status", None) == "READY"

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
    ):
        if self._index is None:
            self._create_index(embeddings.shape[1])

        sample_ids = np.asarray(sample_ids)
        label_ids = np.asarray(label_ids) if label_ids is not None else None

        _sample_ids, _label_ids, ii, _ = fbu.add_ids(
            sample_ids,
            label_ids,
            self._sample_ids,
            self._label_ids,
            patches_field=self.config.patches_field,
            overwrite=overwrite,
            allow_existing=allow_existing,
            warn_existing=warn_existing,
        )

        if ii.size == 0:
            return

        fbu.add_embeddings(
            self._dataset,
            embeddings[ii, :].tolist(),  # MongoDB requires list fields
            sample_ids[ii],
            label_ids[ii] if label_ids is not None else None,
            self.config.embeddings_field,
            patches_field=self.config.patches_field,
        )

        self._sample_ids = _sample_ids
        self._label_ids = _label_ids

        if reload:
            super().reload()

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        _sample_ids, _label_ids, rm_inds = fbu.remove_ids(
            sample_ids,
            label_ids,
            self._sample_ids,
            self._label_ids,
            patches_field=self.config.patches_field,
            allow_missing=allow_missing,
            warn_missing=warn_missing,
        )

        if rm_inds.size == 0:
            return

        if self.config.patches_field is not None:
            rm_sample_ids = None
            rm_label_ids = self._label_ids[rm_inds]
        else:
            rm_sample_ids = self._sample_ids[rm_inds]
            rm_label_ids = None

        fbu.remove_embeddings(
            self._dataset,
            self.config.embeddings_field,
            sample_ids=rm_sample_ids,
            label_ids=rm_label_ids,
            patches_field=self.config.patches_field,
        )

        self._sample_ids = _sample_ids
        self._label_ids = _label_ids

        if reload:
            super().reload()

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        _embeddings, _sample_ids, _label_ids = fbu.get_embeddings(
            self._dataset,
            patches_field=self.config.patches_field,
            embeddings_field=self.config.embeddings_field,
        )

        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

            inds = _get_inds(
                label_ids,
                _label_ids,
                "label",
                allow_missing,
                warn_missing,
            )

            embeddings = _embeddings[inds, :]
            sample_ids = _sample_ids[inds]
            label_ids = np.asarray(label_ids)
        elif sample_ids is not None:
            if etau.is_str(sample_ids):
                sample_ids = [sample_ids]

            if self.config.patches_field is not None:
                sample_ids = set(sample_ids)
                bools = [_id in sample_ids for _id in _sample_ids]
                inds = np.nonzero(bools)[0]
            else:
                inds = _get_inds(
                    sample_ids,
                    _sample_ids,
                    "sample",
                    allow_missing,
                    warn_missing,
                )

            embeddings = _embeddings[inds, :]
            sample_ids = _sample_ids[inds]
            if self.config.patches_field is not None:
                label_ids = _label_ids[inds]
            else:
                label_ids = None
        else:
            embeddings = _embeddings
            sample_ids = _sample_ids
            label_ids = _label_ids

        return embeddings, sample_ids, label_ids

    def reload(self):
        sample_ids, label_ids = self._parse_data(self._dataset, self.config)
        self._sample_ids = sample_ids
        self._label_ids = label_ids

        super().reload()

    def cleanup(self):
        if self._index is None:
            return

        try:
            coll = self._dataset._sample_collection
            coll.drop_search_index(self.config.index_name)
        except OperationFailure:
            # requires MongoDB Atlas 7.0 or later
            pass

        self._index = None

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        if query is None:
            raise ValueError("MongoDB does not support full index neighbors")

        if reverse is True:
            raise ValueError(
                "MongoDB does not support least similarity queries"
            )

        if aggregation not in (None, "mean"):
            raise ValueError(
                f"MongoDB does not support {aggregation} aggregation"
            )

        if k is None:
            k = min(self.index_size, self.config.max_k)

        # https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-stage
        num_candidates = min(10 * k, self.config.max_k)

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        if self.has_view:
            # https://www.mongodb.com/community/forums/t/258494
            logger.warning(
                "The MongoDB backend does not yet support views; the full "
                "index will instead be queried, which may result in fewer "
                "matches in your current view"
            )
            index_ids = None
            # if self.config.patches_field is not None:
            #     index_ids = self.current_label_ids
            # else:
            #     index_ids = self.current_sample_ids
        else:
            index_ids = None

        dataset = self._dataset

        ids = []
        dists = []
        for q in query:
            search = {
                "index": self.config.index_name,
                "path": self.config.embeddings_field,
                "limit": k,
                "numCandidates": num_candidates,
                "queryVector": q.tolist(),
            }

            if index_ids is not None:
                search["filter"] = {
                    "_id": {"$in": [ObjectId(_id) for _id in index_ids]}
                }
            elif dataset.media_type == fom.GROUP:
                # $vectorSearch must be the first stage in all pipelines, so we
                # have to incorporate slice selection as a $filter
                name_field = dataset.group_field + ".name"
                group_slice = self.view.group_slice or dataset.group_slice
                search["filter"] = {name_field: {"$eq": group_slice}}

            project = {"_id": 1}
            if return_dists:
                project["score"] = {"$meta": "vectorSearchScore"}

            # https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-stage
            pipeline = [{"$vectorSearch": search}, {"$project": project}]
            matches = list(
                dataset._aggregate(pipeline=pipeline, manual_group_select=True)
            )

            ids.append([str(m["_id"]) for m in matches])
            if return_dists:
                dists.append([m["score"] for m in matches])

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        embeddings = self._get_embeddings(query_ids)
        num_missing = len(query_ids) - len(embeddings)
        for e in embeddings:
            num_missing += int(e is None)

        if num_missing > 0:
            if single_query:
                raise ValueError("The query ID does not exist in this index")
            else:
                raise ValueError(
                    f"{num_missing} query IDs do not exist in this index"
                )

        query = np.array(embeddings)
        if single_query:
            query = query[0, :]

        return query

    def _get_embeddings(self, query_ids):
        dataset = self._dataset
        patches_field = self.config.patches_field
        embeddings_field = self.config.embeddings_field
        if patches_field is not None:
            _, embeddings_path = dataset._get_label_field_path(
                patches_field, embeddings_field
            )
            view = dataset.filter_labels(
                patches_field, F("_id").is_in(query_ids)
            )
            embeddings = view.values(embeddings_path, unwind=True)
        else:
            view = dataset.select(query_ids)
            embeddings = view.values(embeddings_field)

        return embeddings

    @staticmethod
    def _parse_data(samples, config):
        if config.patches_field is not None:
            samples = samples.filter_labels(
                config.patches_field, F(config.embeddings_field).exists()
            )
        else:
            samples = samples.exists(config.embeddings_field)

        return fbu.get_ids(samples, patches_field=config.patches_field)

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)


def _get_inds(ids, index_ids, ftype, allow_missing, warn_missing):
    if etau.is_str(ids):
        ids = [ids]

    ids_map = {_id: i for i, _id in enumerate(index_ids)}

    inds = []
    bad_ids = []

    for _id in ids:
        idx = ids_map.get(_id, None)
        if idx is not None:
            inds.append(idx)
        else:
            bad_ids.append(_id)

    num_missing = len(bad_ids)

    if num_missing > 0:
        if not allow_missing:
            raise ValueError(
                "Found %d %s IDs (eg '%s') that are not present in the index"
                % (num_missing, ftype, bad_ids[0])
            )

        if warn_missing:
            logger.warning(
                "Ignoring %d %s IDs that are not present in the index",
                num_missing,
                ftype,
            )

    return np.array(inds)
